import numpy as np
import pandas as pd

class AnnotationsHelper:
    """This help to split annotations in .txt format.
    
    Args:
        annotations_path: a path with a .txt annotations file.
    """
    def __init__(self, annotations_path):
        with open(annotations_path, 'r') as f:
            txt = f.readlines()
            self.annotations = [line.strip() for line in txt if len(line.strip().split()[1:]) != 0]
        self.annotations_path = annotations_path

    def shuffle(self):
        """shuffle annotations"""
        temp_data = np.asarray(self.annotations)
        temp_data = pd.DataFrame(temp_data)
        self.annotations = temp_data.sample(frac=1)
        self.annotations = self.annotations.values.tolist()
    
    def export(self, data, file_path):
        """To export a dataframe like a .txt file.

        Args:
            data: a pandas dataframe
            file_path: output file path
        """
        data.to_csv(file_path, header=None, index=None, sep='\t')
    
    def split(self, train_percentage=0.8, random_state=25):
        """Split annotations in two dataframes in reference of a percentage.

        Args:
            train_percentage: a float between (0, 1) that corresponds with train data proportion.
            random_state: int, array-like, BitGenerator, np.random.RandomState

        Returns:
            A tuple where the first element is train data (DataFrame) and the second is test data (DataFrame)
        """
        data = np.asarray(self.annotations)
        data = pd.DataFrame(data)
        train_data = data.sample(frac=train_percentage, random_state=random_state)
        test_data = data.drop(train_data.index)

        return train_data, test_data



from logging import exception
import tensorflow as tf

from tensorflow.keras import Input 

from py2vision.models.layers.conv2d_bn_leaky_relu_layer import conv2d_bn_leaky_relu_layer
from py2vision.models.layers.upsample_layer import UpsampleLayer


class BuildYoloV3Tiny(tf.keras.Model):
    """Build YoloV3 tiny model given a backbone

    Args:
        name: a string with the name of the model
        backbone: an object with a backbone network
        num_class: an integer with the quantity of classes 
        
    Returns:
        A list where the first one is used to predict large-sized objects, 
        the second one is used to predict medium-sized objects and the  last one is the input shape returned.
    """
    def __init__(self, backbone, num_class):
        super().__init__()
        if not isinstance(num_class, int):
            raise ValueError('num_class has to be an integer')
        self.base_outputs = backbone
        self.num_class = num_class
        
    def call(self, input_shape):
        if len(input_shape) != 3: raise ValueError("input shape should have a len == 3")
        input_shape = Input([input_shape[0], input_shape[1], input_shape[2]])
        try:
            route_1, x = self.base_outputs.build_model(input_shape)
        except ValueError:
            raise Exception('Backbone output shape mismatch with yolov3 tiny input shape')

        x = conv2d_bn_leaky_relu_layer(x, (1, 1, 1024, 256))

        conv_lobj_branch = conv2d_bn_leaky_relu_layer(x, (3, 3, 256, 512))
        # conv_lbbox is used to predict large-sized objects , Shape = [None, 13, 13, 255] (if num_class == 80 like coco dataset) 
        conv_lbbox = conv2d_bn_leaky_relu_layer(conv_lobj_branch, (1, 1, 512, 3*(self.num_class + 5)), activate=False, bn=False)
                        
        x = conv2d_bn_leaky_relu_layer(x, (1, 1, 256, 128))
        # upsample here uses the nearest neighbor interpolation method, which has the advantage that the
        # upsampling process does not need to learn, thereby reducing the network parameter  
        x = UpsampleLayer()(x)
        x = tf.concat([x, route_1], axis=-1)

        conv_mobj_branch = conv2d_bn_leaky_relu_layer(x, (3, 3, 128, 256))
        # conv_mbbox is used to predict medium size objects, shape = [None, 26, 26, 255] (if num_class == 80 like coco dataset) 
        conv_mbbox = conv2d_bn_leaky_relu_layer(conv_mobj_branch, (1, 1, 256, 3*(self.num_class + 5)), activate=False, bn=False)

        return [conv_mbbox, conv_lbbox, input_shape]


        








import cv2 as cv

def re_projection_error(objpoints, rvecs, tvecs, mtx, imgpoints, dist):
    """Calculates the error when calibrating a camera. 
    when the error approaches zero, 
    the more accurate are the parameters found. 
    Given the intrinsic, distortion, rotation and translation matrices, we must first transform the 
    translation matrices, we must first transform the object point to the point of the camera. 
    the object point to image point using cv.projectPoints().
    Then, we will calculate the absolute norm between what we obtained with our transformation and the algorithm. 
    obtained with our transformation and the corner search algorithm. 
    To find the average error, we calculate the arithmetic average
    of the errors calculated for all calibration images.
    
    Args:
        objpoints: array of object points expressed in the world coordinate frame. A 1-channel 3xN/Nx3 or 3-channel 1xN/Nx1, where N is the number of points in the view.
        rvecs: the rotation vector (Rodrigues) which, together with tvec, performs a change of base from the world coordinate system to the camera, see calibrateCamera for more details.
        tvecs: The translation vector, see the description of the parameters above.
        mtx: The intrinsic matrix of the camera.
        imgpoints: reference points to compare with the reprojected objpoints.
        dist: Input vector of the distortion coefficients.

    Returns
        a float with the mean error with L2 norm.
    """
    mean_error = 0
    for i in range(len(objpoints)):
        imgpoints2, _ = cv.projectPoints(objpoints[i], rvecs[i], tvecs[i], mtx, dist)
        error = cv.norm(imgpoints[i], imgpoints2, cv.NORM_L2)/len(imgpoints2)
        mean_error += error
    return mean_error/len(objpoints)


import cv2 as cv
import numpy as np
import time
import os

from py2vision.input_output.camera import Camera
from py2vision.recognition.detection_mode import DetectRealTime
from py2vision.recognition.detection_mode import DetectImage
from py2vision.stereo.match_method import Matcher
from py2vision.stereo.stereo_builder import StereoController
from py2vision.stereo.standard_stereo import StandardStereoBuilder

class VisionSystem:
    """Provide an interface to apply recognition and stereo vision. Initialization of all necessary parameteres to implement an stereo-recognition system 

    Args:
        cam_left: a camera instance which can be streaming, mp4 file path, image path or even realtime source.
        cam_right: a camera instance which can be streaming, mp4 file path, image path or even realtime source.
        stereo_maps_path: a path stereo maps with rectify images.
        q_matrix: a 4x4 array with the following structure, [[1 0   0          -cx     ][0 1   0          -cy     ][0 0   0           f      ][0 0 -1/Tx (cx - cx')/Tx ]]
            cx: is the principal point x in left image
            cx': is the principal point x in right image
            cy: is the principal point y in left image
            f: is the focal lenth in left image
            Tx: The x coordinate in Translation matrix
    """
    def __init__(self, cam_left: Camera, cam_right: Camera, stereo_maps_path, matcher: Matcher, q_matrix):
        self.camL = cam_left
        self.camR = cam_right
        builder = StandardStereoBuilder(self.camL, self.camR, stereo_maps_path)
        self.stereo_controller = StereoController()
        self.stereo_controller.stereo_builder = builder
        self.matcher = matcher
        self.Q = q_matrix

    def realtime_or_video_pipeline(self, model, class_file_name, output_path="", input_size=416, 
                score_threshold=0.3, iou_threshold=0.45, rectangle_colors='', nms_method="nms", 
                post_process_match=True, lmbda=128.0, sigma=1.5, downsample_for_match=2, show_window=True, otsu_thresh_inverse=True, text_colors=(255,255,0)):
        """ Implement a stereo recognition system for video or streaming 

        Args:
            model: expects a tensorflow model trained.
            class_file_name: it's the path of classes .txt file
            output_path: if is an empty string, it won't be saved, but it is a path it save like a video.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.
            post_process_match: if is true apply post_process and return an improved disparity map, otherwise return left disparity map without post processing.
            lmbda: is a parameter defining the amount of regularization during filtering. Larger values force filtered disparity map edges to adhere more to source image edges. Typical value is 8000. Only valid in post processing step
            sigma: is a parameter defining how sensitive the filtering process is to source image edges. Large values can lead to disparity leakage through low-contrast edges. Small values can make the filter too sensitive to noise and textures in the source image. Typical values range from 0.8 to 2.0. Only valid in post processing step.
            downsample_for_match: if true, will apply the blur on both frames and demultiply it. The downsampling factor can be 2, 4, 8, 16, 32, 64. If the downsample factor is 1 or None or False it will not apply the downsampling.
            show_window: shows a window with the application.
            otsu_thresh_inverse: The Otsu threshold transforms a grayscale image into a binary image, if this variable is True the binary image will favor darker pixels otherwise it will favor lighter pixels.
            text_colors: a tuple that represents (R, G, B) colors for drawed text.
        """
        times, times_2 = [], []
        detector = DetectRealTime()
        inputL = detector.camera_input(self.camL)
        inputR = detector.camera_input(self.camR)
        if isinstance(inputL, cv.VideoCapture):
            out, inputL, _ = detector.prepare_input(inputL, output_path)

        while True:
            if isinstance(inputL, cv.VideoCapture):
                retL, imgL = inputL.read()
                retR, imgR = inputR.read()
                if not retL: break
                if not retR: break
            else: 
                imgL = self.camera_input(self.camL)
                imgR = self.camera_input(self.camR)
            try:
                frameL = cv.cvtColor(imgL, cv.COLOR_BGR2RGB)
                frameL = cv.cvtColor(frameL, cv.COLOR_BGR2RGB)
            except:
                break
            # Preprocess
            left_for_matcher, right_for_matcher = self.stereo_controller.pre_process_step(imgL, imgR, downsample_for_match)
            left_for_detector = detector.pre_process(frameL, input_size)
            # match and predict
            t1 = time.time()
            left_disp, right_disp, left_matcher = self.stereo_controller.stereo_builder.match(left_for_matcher, 
                                                                                                right_for_matcher, self.matcher, False)
            pred_bbox = detector.predict(model, left_for_detector)
            t2 = time.time()
            # post_process
            if post_process_match:
                disparity = self.stereo_controller.stereo_builder.post_process(left_for_matcher, left_disp, right_disp, left_matcher, lmbda=lmbda, sigma=sigma, metrics=False)
            else: 
                disparity = left_disp
            # recover original size
            if downsample_for_match in [1, None, False]:
                    n_upsamples = 0
            else:
                n_upsamples = [2**p for p in range(1, 7)].index(downsample_for_match)
                n_upsamples += 1
            if n_upsamples > 0:
                for i in range(n_upsamples):
                    disparity = cv.pyrUp(disparity)
            # resize bboxes
            bboxes = detector.postprocess_boxes(disparity, pred_bbox, input_size,
                                             score_threshold, iou_threshold, nms_method)
            # get distance
            average_homogeneous_points = []
            for bbox in bboxes: 
                blur = cv.GaussianBlur(imgL[int(bbox[1]):int(bbox[3]), int(bbox[0]):int(bbox[2])],(5,5),0)
                blur = cv.cvtColor(blur,cv.COLOR_BGR2GRAY)
                if otsu_thresh_inverse:
                    _, mask = cv.threshold(blur,0,255, cv.THRESH_BINARY_INV+cv.THRESH_OTSU)
                else:
                    _, mask = cv.threshold(blur,0,255, cv.THRESH_BINARY+cv.THRESH_OTSU)
                mask_index = np.argwhere(mask)
                mask_index[:, [1, 0]] = mask_index[:, [0, 1]]
                mask_index += np.array([int(bbox[0]), int(bbox[1])])
                points_3D = self.stereo_controller.stereo_builder.estimate_3D_points(mask_index, disparity, self.Q)
                points_3D = np.asarray(points_3D)
                average_homogeneous_points.append(np.mean(points_3D, axis=0))

            # divide by W
            if len(average_homogeneous_points) > 0:
                average_homogeneous_points = np.asarray(average_homogeneous_points)
                last_w = average_homogeneous_points[:, -1]
                average_homogeneous_points = average_homogeneous_points/last_w[:, None]
                average_homogeneous_points = list(average_homogeneous_points)
                
            t3 = time.time()
            times.append(t2-t1)
            times_2.append(t3-t1)
            times = times[-20:]
            times_2 = times_2[-20:]
            
            ms = sum(times)/len(times)*1000
            fps = 1000 / ms
            fps2 = 1000 / (sum(times_2)/len(times_2)*1000)
            # draw on image
            print("Time: {:.2f}ms, Detection FPS: {:.1f}, total FPS: {:.1f}".format(ms, fps, fps2))
            frame = detector.draw(frameL, bboxes, class_file_name, rectangle_colors, homogeneous_points=average_homogeneous_points, text_colors=text_colors)

            cv.putText(frame, "Time: {:.1f}FPS".format(fps), (0, 30),
                          cv.FONT_HERSHEY_COMPLEX_SMALL, 1, (0, 0, 255), 2)
            if output_path != '': out.write(frame)
            if show_window:
                detector.show(frame, self.camL)
            if cv.waitKey(25) & 0xFF == ord("q"):
                cv.destroyAllWindows()
                break
        cv.destroyAllWindows()

    def image_pipeline(self, model, class_file_name, output_path="", input_size=416, 
                score_threshold=0.3, iou_threshold=0.45, rectangle_colors='', nms_method="nms", 
                post_process_match=True, lmbda=128.0, sigma=1.5, downsample_for_match=2, show_window=True, otsu_thresh_inverse=True, text_colors=(255,255,0)):
        """ Implement a stereo recognition system for images.

        Args:
            model: expects a tensorflow model trained.
            class_file_name: it's the path of classes .txt file
            output_path: if is an empty string, it won't be saved, but it is a path it save like a video.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.
            post_process_match: if is true apply post_process and return an improved disparity map, otherwise return left disparity map without post processing.
            lmbda: is a parameter defining the amount of regularization during filtering. Larger values force filtered disparity map edges to adhere more to source image edges. Typical value is 8000. Only valid in post processing step
            sigma: is a parameter defining how sensitive the filtering process is to source image edges. Large values can lead to disparity leakage through low-contrast edges. Small values can make the filter too sensitive to noise and textures in the source image. Typical values range from 0.8 to 2.0. Only valid in post processing step.
            downsample_for_match: if true, will apply the blur on both frames and demultiply it. The downsampling factor can be 2, 4, 8, 16, 32, 64. If the downsample factor is 1 or None or False it will not apply the downsampling.
            show_window: shows a window with the application.
            otsu_thresh_inverse: The Otsu threshold transforms a grayscale image into a binary image, if this variable is True the binary image will favor darker pixels otherwise it will favor lighter pixels.
            text_colors: a tuple that represents (R, G, B) colors for drawed text.

        Returns:
            an image processed, bounding boxes and 3D points.

        Raises:
            ValueError: if input images are not some of this formats [".bmp", ".dib", ".jpg", ".jpeg", ".jpe", ".png", ".webp"].
        """
        detector = DetectImage()
        compatible_outputs = [".bmp", ".dib", ".jpg", ".jpeg", ".jpe", ".png", ".webp"]
        if not os.path.splitext(output_path)[1] in compatible_outputs: 
            raise ValueError("output_path only can be one of this: {}".format(compatible_outputs))
        imgL = detector.prepare_input(self.camL.source)
        imgL_for_matcher = cv.imread(self.camL.source)
        imgR_for_matcher = cv.imread(self.camR.source)
        # pre-process for detect
        left_for_detector = detector.pre_process(imgL, input_size)
        # compute disparity and predict
        disparity, _ = self.stereo_controller.compute_disparity(imgL_for_matcher, imgR_for_matcher, self.matcher, downsample_for_match, lmbda, sigma, post_process_match, False)
        pred_bbox = detector.predict(model, left_for_detector)
        # resize bboxes
        bboxes = detector.postprocess_boxes(disparity, pred_bbox, input_size,
                                             score_threshold, iou_threshold, nms_method)
        # get distance
        average_homogeneous_points = []
        for bbox in bboxes: 
            blur = cv.GaussianBlur(imgL[int(bbox[1]):int(bbox[3]), int(bbox[0]):int(bbox[2])],(5,5),0)
            blur = cv.cvtColor(blur,cv.COLOR_BGR2GRAY)
            if otsu_thresh_inverse:
                _, mask = cv.threshold(blur,0,255, cv.THRESH_BINARY_INV+cv.THRESH_OTSU)
            else:
                _, mask = cv.threshold(blur,0,255, cv.THRESH_BINARY+cv.THRESH_OTSU)
            mask_index = np.argwhere(mask)
            mask_index[:, [1, 0]] = mask_index[:, [0, 1]]
            mask_index += np.array([int(bbox[0]), int(bbox[1])])
            points_3D = self.stereo_controller.stereo_builder.estimate_3D_points(mask_index, disparity, self.Q)
            points_3D = np.asarray(points_3D)
            average_homogeneous_points.append(np.mean(points_3D, axis=0))

        # divide by W
        if len(average_homogeneous_points) > 0:
            average_homogeneous_points = np.asarray(average_homogeneous_points)
            last_w = average_homogeneous_points[:, -1]
            average_homogeneous_points = average_homogeneous_points/last_w[:, None]
            average_homogeneous_points = list(average_homogeneous_points)
        # draw on image
        frame = detector.draw(imgL, bboxes, class_file_name, rectangle_colors, homogeneous_points=average_homogeneous_points, text_colors=text_colors)
        if output_path != '': cv.imwrite(output_path, frame)
        if show_window:
            # Show the image
            cv.imshow(self.camL.id, frame)
            # Load and hold the image
            cv.waitKey(0)
            # To close the window after the required kill value was provided
            cv.destroyAllWindows()

        return frame, bboxes, average_homogeneous_points



import numpy as np
import cv2 as cv
import numpy as np
import colorsys
import random
import cv2 as cv

from py2vision.utils.label_utils import read_class_names

def draw_lines(img1,img2,lines,pts1,pts2):
    ''' img1 - image on which we draw the epilines for the points in img2 lines - corresponding epilines '''
    r,c = img1.shape
    img1 = cv.cvtColor(img1,cv.COLOR_GRAY2BGR)
    img2 = cv.cvtColor(img2,cv.COLOR_GRAY2BGR)
    for r,pt1,pt2 in zip(lines,pts1,pts2):
        color = tuple(np.random.randint(0,255,3).tolist())
        x0,y0 = map(int, [0, -r[2]/r[1] ])
        x1,y1 = map(int, [c, -(r[2]+r[0]*c)/r[1] ])
        img1 = cv.line(img1, (x0,y0), (x1,y1), color,1)
        img1 = cv.circle(img1,tuple(pt1),5,color,-1)
        img2 = cv.circle(img2,tuple(pt2),5,color,-1)
    return img1,img2

def draw_bbox(image, bboxes, class_file_name, show_label=True, show_confidence = True, text_colors=(255,255,0), rectangle_colors='', tracking=False, homogeneous_points=None):   
    """Draw bounding boxes on images

    Args:
        image: an array which correspond with an image
        bboxes: their bounding boxes.
        class_file_name: a path with a .txt file where the classes are saved.
        show_label: a boolean to show or hide object label.
        show_confidence: a boolean to show or hide confidence level.
        text_colors: a tuple that represents (R, G, B) colors.
        rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
        homogeneous_points: an array with dimensions n x 4 where each row is like (X, Y, Z, W). However if is None it won't be drawed.
        
    Returns:
        An image with bounding boxes and homogeneous coordinates.
    """
    classes = read_class_names(class_file_name)
    num_classes = len(classes)
    image_h, image_w, _ = image.shape
    hsv_tuples = [(1.0 * x / num_classes, 1., 1.) for x in range(num_classes)]
    #print("hsv_tuples", hsv_tuples)
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), colors))

    random.seed(0)
    random.shuffle(colors)
    random.seed(None)

    for i, bbox in enumerate(bboxes):
        coor = np.array(bbox[:4], dtype=np.int32)
        score = bbox[4]
        class_ind = int(bbox[5])
        bbox_color = rectangle_colors if rectangle_colors != '' else colors[class_ind]
        bbox_thick = int(0.6 * (image_h + image_w) / 1000)
        if bbox_thick < 1: bbox_thick = 1
        fontScale = 0.75 * bbox_thick
        (x1, y1), (x2, y2) = (coor[0], coor[1]), (coor[2], coor[3])

        # put object rectangle
        cv.rectangle(image, (x1, y1), (x2, y2), bbox_color, bbox_thick*2)

        if show_label:
            # get text label
            score_str = " {:.2f}".format(score) if show_confidence else ""

            if tracking: score_str = " "+str(score)

            label = "{}".format(classes[class_ind]) + score_str

            # get text size
            (text_width, text_height), baseline = cv.getTextSize(label, cv.FONT_HERSHEY_COMPLEX_SMALL,
                                                                  fontScale, thickness=bbox_thick)
            # put filled text rectangle
            cv.rectangle(image, (x1, y1), (x1 + text_width, y1 - text_height - baseline), bbox_color, thickness=cv.FILLED)

            # put text above rectangle
            cv.putText(image, label, (x1, y1-4), cv.FONT_HERSHEY_COMPLEX_SMALL,
                        fontScale, text_colors, bbox_thick, lineType=cv.LINE_AA)

        if homogeneous_points != None:
            y_position = y1 + 14
            coordinates = ["X", "Y:", "Z", "W"]
            for num, string_coor in enumerate(coordinates):
                coor_label = "{}: {:.2f}".format(string_coor, homogeneous_points[i][num])
                cv.putText(image, coor_label, (x1 + 3, y_position), cv.FONT_HERSHEY_COMPLEX_SMALL, fontScale, text_colors, bbox_thick, lineType=cv.LINE_AA)
                y_position += 14


    return image


from __future__ import annotations
from abc import ABC, abstractmethod

import cv2 as cv


class Matcher():
    """
    The Context defines the interface of interest to clients.
    The matcher accepts a strategy through the constructor, but also provides a setter to change it at runtime.
    """

    def __init__(self, strategy: MatcherStrategy) -> None:
        self._strategy = strategy

    @property
    def strategy(self) -> MatcherStrategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """

        return self._strategy

    @strategy.setter
    def strategy(self, strategy: MatcherStrategy) -> None:
        """
        Usually, the Context allows replacing a Strategy object at runtime.
        """

        self._strategy = strategy

    def match(self):
        """
        The Context delegates some work to the Strategy object instead of
        implementing multiple versions of the algorithm on its own.
        """
        return self._strategy.match()



class MatcherStrategy(ABC):
    """
    The Strategy interface declares operations common to all supported versions
    of some algorithm.

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.
    """

    @abstractmethod
    def match(self):
        pass

class StereoSGBM(MatcherStrategy):
    """ To create an instance of stereo SGBM algorithm.

    Args:
        min_disp: Minimum possible disparity value. Normally, it is zero but sometimes rectification algorithms can shift images, so this parameter needs to be adjusted accordingly.
        max_disp: Maximum disparity minus minimum disparity. The value is always greater than zero. In the current implementation, this parameter must be divisible by 16.
        window_size: Matched block size. It must be an odd number >=1 . Normally, it should be somewhere in the 3..11 range.
        p1: The first parameter controlling the disparity smoothness. 
        p2: The second parameter controlling the disparity smoothness. The larger the values are, the smoother the disparity is. P1 is the penalty on the disparity change by plus or minus 1 between neighbor pixels. P2 is the penalty on the disparity change by more than 1 between neighbor pixels. The algorithm requires P2 > P1 . See stereo_match.cpp sample where some reasonably good P1 and P2 values are shown (like 8*number_of_image_channels*SADWindowSize*SADWindowSize and 32*number_of_image_channels*SADWindowSize*SADWindowSize, respectively).
        pre_filter_cap: Truncation value for the prefiltered image pixels. The algorithm first computes x-derivative at each pixel and clips its value by [-preFilterCap, preFilterCap] interval. The result values are passed to the Birchfield-Tomasi pixel cost function.
        mode: Set it to StereoSGBM_MODE_HH to run the full-scale two-pass dynamic programming algorithm. It will consume O(W*H*numDisparities) bytes, which is large for 640x480 stereo and huge for HD-size pictures. By default, it is set to false .
        speckle_window_size: Maximum size of smooth disparity regions to consider their noise speckles and invalidate. Set it to 0 to disable speckle filtering. Otherwise, set it somewhere in the 50-200 range.
        speckle_range: Maximum disparity variation within each connected component. If you do speckle filtering, set the parameter to a positive value, it will be implicitly multiplied by 16. Normally, 1 or 2 is good enough.
        uniqueness_ratio: Margin in percentage by which the best (minimum) computed cost function value should "win" the second best value to consider the found match correct. Normally, a value within the 5-15 range is good enough.
        disp_12_max_diff: Maximum allowed difference (in integer pixel units) in the left-right disparity check. Set it to a non-positive value to disable the check.
    """
    def __init__(self, min_disp=0, max_disp=160, window_size=3, p1=24*3*3, p2=96*3*3, pre_filter_cap=63, mode=cv.StereoSGBM_MODE_HH, speckle_window_size=1100, speckle_range=1, uniqueness_ratio=5, disp_12_max_diff=-1):
        try:
            if max_disp <= 0 or max_disp % 16 != 0:
                raise ValueError
        except ValueError:
            print("Incorrect max_disparity value: it should be positive and divisible by 16")
        try:
            if window_size <= 0 or window_size % 2 != 1:
                raise ValueError
        except ValueError:
            print("Incorrect window_size value: it should be positive and odd")
        max_disp /= 2
        if(max_disp % 16 != 0):
            max_disp += 16-(max_disp % 16)

        self.min_disp = min_disp
        self.max_disp = max_disp
        self.window_size = window_size
        self.p1 = p1
        self.p2 = p2
        self.pre_filter_cap = pre_filter_cap
        self.mode = mode
        self.speckle_window_size = speckle_window_size
        self.speckle_range = speckle_range
        self.uniqueness_ratio = uniqueness_ratio
        self.disp_12_max_diff = disp_12_max_diff

    def match(self):
        """ Return stereo sgbm instance """
        return cv.StereoSGBM_create(self.min_disp, int(self.max_disp), self.window_size, self.p1, self.p2, self.disp_12_max_diff, self.pre_filter_cap, self.uniqueness_ratio, self.speckle_window_size, self.speckle_range, self.mode)


from __future__ import annotations
from abc import ABC, abstractmethod


from py2vision.stereo.match_method import Matcher

import cv2 as cv

class StereoSystemBuilder(ABC):
    """
    The Builder interface specifies methods for stereo contoller
    """

    @property
    @abstractmethod
    def get_product(self):
        pass

    @abstractmethod
    def pre_process(self):
        pass

    @abstractmethod
    def match(self):
        pass

    @abstractmethod
    def post_process(self):
        pass
    
    @abstractmethod
    def estimate_depth_map(self):
        pass

    @abstractmethod
    def estimate_3D_points(self):
        pass


class StereoController:
    """
    The StereoController is only responsible for executing the stereo steps in a particular sequence. It is helpful when producing products according to a specific order or configuration.
    """

    def __init__(self) -> None:
        self._stereo_builder = None

    @property
    def stereo_builder(self) -> StereoSystemBuilder:
        return self._stereo_builder

    @stereo_builder.setter
    def stereo_builder(self, builder: StereoSystemBuilder) -> None:
        """
        To change stereo builder instance used by this class
        """
        self._stereo_builder = builder

    def pre_process_step(self, frameL, frameR, downsample=2):
        """ First, it transform from BGR to gray, next apply rectification and finally apply pyramid subsampling.

        Args: 
            frameL: it's the left frame
            frameR: it's the right frame
            downsample: if it is true, it will apply blurry in both frames and downsamples it. The downsampling factor just can be 2, 4, 8, 16, 32, 64. If downsample factor is 1 or None or False won't apply downsampling.

        Returns:
            Both frames to apply stereo correspondence or apply more processing to the images.
        """
        left_for_matcher, right_for_matcher = self.stereo_builder.pre_process(frameL, frameR, downsample)
        return left_for_matcher, right_for_matcher
    
    def get_epilines(self, frameL, frameR):
        """ Draw epilines to both frames.

        Args: 
            frameL (arr): it's the left frame.
            frameR (arr): it's the right frame.

        Returns:
            Two elements, the left and right frame with epilines.
        """
        left_epilines, right_epilines = self.stereo_builder.find_epilines(frameL, frameR)
        return left_epilines, right_epilines
        
    def compute_disparity(self, frameL, frameR, matcher: Matcher, downsample=2, lmbda=8000, sigma=1.5, post_process=True, metrics=True):
        """ Apply the pre process step, next compute left and right disparity maps, and finally execute the post process with a wls filter to improve the final result.

        Args: 
            frameL: it's the left frame.
            frameR: it's the right frame.
            matcher: A Matcher instance.
            downsample: if it is true, it will apply blurry in both frames and downsamples it. The downsampling factor just can be 2, 4, 8, 16, 32, 64. If downsample factor is 1 or None or False won't apply downsampling.
            lmbda: is a parameter defining the amount of regularization during filtering. Larger values force filtered disparity map edges to adhere more to source image edges. Typical value is 8000. Only valid in post processing step
            sigma: is a parameter defining how sensitive the filtering process is to source image edges. Large values can lead to disparity leakage through low-contrast edges. Small values can make the filter too sensitive to noise and textures in the source image. Typical values range from 0.8 to 2.0. Only valid in post processing step
            post_process: if is true apply post_process and return improved disparity map, otherwise return left disparity map without post processing.
            metrics: if is true print by console the time of execution of correspondence and post process steps.

        Returns:
            Two elements, disparity map and correspondence method used.
        """
        left_for_matcher, right_for_matcher = self.stereo_builder.pre_process(frameL, frameR, downsample)
        left_disp, right_disp, left_matcher = self.stereo_builder.match(left_for_matcher, right_for_matcher, matcher, metrics=metrics)
        if post_process:
            disparity = self.stereo_builder.post_process(left_for_matcher, left_disp, right_disp, left_matcher, lmbda=lmbda, sigma=sigma, metrics=metrics)
        else:
            disparity = left_disp
        # recover original size
        if downsample in [1, None, False]:
                n_upsamples = 0
        else:
            n_upsamples = [2**p for p in range(1, 7)].index(downsample)
            n_upsamples += 1
        if n_upsamples > 0:
            for i in range(n_upsamples):
                disparity = cv.pyrUp(disparity)

        return disparity, left_matcher

    def compute_disparity_color_map(self, frameL, frameR, matcher: Matcher, downsample=2, lmbda=8000, sigma=1.5, post_process=True, metrics=True):
        """ Apply the pre process step, next compute left and right disparity maps, then execute the post process with a wls filter to improve the final result, and finally compute disparity map with color.

        Args: 
            frameL: it's the left frame.
            frameR: it's the right frame.
            matcher: A Matcher instance.
            downsample: if it is true, it will apply blurry in both frames and downsamples it. The downsampling factor just can be 2, 4, 8, 16, 32, 64. If downsample factor is 1 or None or False won't apply downsampling.
            lmbda: is a parameter defining the amount of regularization during filtering. Larger values force filtered disparity map edges to adhere more to source image edges. Typical value is 8000. Only valid in post processing step
            sigma: is a parameter defining how sensitive the filtering process is to source image edges. Large values can lead to disparity leakage through low-contrast edges. Small values can make the filter too sensitive to noise and textures in the source image. Typical values range from 0.8 to 2.0. Only valid in post processing step.
            post_process: if is true apply post_process and return improved disparity map, otherwise return left disparity map without post processing.
            metrics: if is true print by console the time of execution of correspondence and post process steps.

        Returns:
            Two elements, disparity map (with 3 colors channels) and correspondence method used
        """
        disparity, matcher = self.compute_disparity(frameL, frameR, matcher, downsample, float(lmbda), sigma, post_process, metrics)
        disparity_colormap = self.stereo_builder.estimate_disparity_colormap(disparity)
        return disparity_colormap, matcher


    def compute_3D_map(self, frameL, frameR, Q, matcher: Matcher, downsample=2, lmbda=8000, sigma=1.5, post_process=True,  metrics=True):
        """ Apply the pre process step, next compute left and right disparity maps, then execute the post process with a wls filter to improve the final result, and finally compute depth map.

        Args: 
            frameL: it's the left frame.
            frameR: it's the right frame.
            Q: a 4x4 array with the following structure, [[1 0   0          -cx     ][0 1   0          -cy     ][0 0   0           f      ][0 0 -1/Tx (cx - cx')/Tx ]], cx: is the principal point x in left image, cx': is the principal point x in right image, cy: is the principal point y in left image, f: is the focal lenth in left image, Tx: The x coordinate in Translation matrix.  
            matcher: A Matcher instance.
            downsample: if it is true, it will apply blurry in both frames and downsamples it. The downsampling factor just can be 2, 4, 8, 16, 32, 64. If downsample factor is 1 or None or False won't apply downsampling.
            lmbda: is a parameter defining the amount of regularization during filtering. Larger values force filtered disparity map edges to adhere more to source image edges. Typical value is 8000. Only valid in post processing step
            sigma: is a parameter defining how sensitive the filtering process is to source image edges. Large values can lead to disparity leakage through low-contrast edges. Small values can make the filter too sensitive to noise and textures in the source image. Typical values range from 0.8 to 2.0. Only valid in post processing step
            post_process: if is true apply post_process and return improved disparity map, otherwise return left disparity map without post processing.
            metrics: if is true print by console the time of execution of correspondence, post process and depth map step.

        Returns:
            reprojected points image, a depth map in RGB and correspondence method used
        """
        disparity, matcher = self.compute_disparity(frameL, frameR, matcher, downsample, float(lmbda), sigma, post_process, metrics)
        return self.stereo_builder.estimate_depth_map(disparity, Q, frameL, metrics=metrics), matcher

    def compute_3D_points(self, frameL, frameR, points, Q, matcher: Matcher, downsample=2, lmbda=8000, sigma=1.5, post_process=True,  metrics=True):
        """ Apply the pre process step, next compute left and right disparity maps, then execute the post process with a wls filter to improve the final result, and finally compute 3D points for an input array that can be [[x1, y1], [x2, y2], [x3, y3], ... [xn, yn]].

        Args: 
            frameL: it's the left frame.
            frameR: it's the right frame.
            points: contains the (x, y) coordinates in image plane to convert to (X, Y, Z)
            Q: a 4x4 array with the following structure, [[1 0   0          -cx     ][0 1   0          -cy     ][0 0   0           f      ][0 0 -1/Tx (cx - cx')/Tx ]], cx: is the principal point x in left image, cx': is the principal point x in right image, cy: is the principal point y in left image, f: is the focal lenth in left image, Tx: The x coordinate in Translation matrix.  
            matcher: A Matcher instance.
            downsample: if it is true, it will apply blurry in both frames and downsamples it. The downsampling factor just can be 2, 4, 8, 16, 32, 64. If downsample factor is 1 or None or False won't apply downsampling.
            lmbda: is a parameter defining the amount of regularization during filtering. Larger values force filtered disparity map edges to adhere more to source image edges. Typical value is 8000. Only valid in post processing step
            sigma: is a parameter defining how sensitive the filtering process is to source image edges. Large values can lead to disparity leakage through low-contrast edges. Small values can make the filter too sensitive to noise and textures in the source image. Typical values range from 0.8 to 2.0. Only valid in post processing step
            post_process: if is true apply post_process and return improved disparity map, otherwise return left disparity map without post processing.
            metrics: if is true print by console the time of execution of correspondence, post process and depth map step.
            
        Returns:
            Two elements, an array of points in 3D homogeneous coordinates (X, Y, Z, W) and correspondence method used
        """
        disparity, matcher = self.compute_disparity(frameL, frameR, matcher, downsample, float(lmbda), sigma, post_process, metrics)
        return self.stereo_builder.estimate_3D_points(points, disparity, Q), matcher


import cv2 as cv
import numpy as np

from py2vision.image_process.frame_decorator import FrameDecorator

class ResizeWithBBox(FrameDecorator):
    def apply(self, target_size, gt_boxes=None):
        """Apply resizing on an image and their bounding boxes.
        
        Args:
             target_size: a tuple or list with the new dimensions of an image.
             gt_boxes: bounding boxes. 
        
        Returns:
            a resized image and their reescaling bounding boxes.
        """
        ih, iw    = target_size
        h,  w, _  = self._frame.apply().shape

        scale = min(iw/w, ih/h)
        nw, nh  = int(scale * w), int(scale * h)
        image_resized = cv.resize(self._frame.apply(), (nw, nh))

        image_paded = np.full(shape=[ih, iw, 3], fill_value=128.0)
        dw, dh = (iw - nw) // 2, (ih-nh) // 2
        image_paded[dh:nh+dh, dw:nw+dw, :] = image_resized
        image_paded = image_paded / 255.

        if gt_boxes is None:
            return image_paded

        else:
            gt_boxes[:, [0, 2]] = gt_boxes[:, [0, 2]] * scale + dw
            gt_boxes[:, [1, 3]] = gt_boxes[:, [1, 3]] * scale + dh
            return image_paded, gt_boxes


class Frame():
    """The base Component interface defines operations that can be altered by decorators.
    
    Attributes:
        img: an image array
    """
    def __init__(self, input_image):
        
        self.img = input_image
        
    def apply(self):
        """To get input image"""
        return self.img

class FrameDecorator(Frame):
    _frame: Frame = None

    def __init__(self, frame: Frame):
        """ The base Decorator class follows the same interface as the other components. The primary purpose of this class is to define the wrapping interface for all concrete decorators. The default implementation of the wrapping code might include a field for storing a wrapped component and the means to initialize it.
        
        Attributes:
            _frame: a Frame instance
        """
        if isinstance(frame, Frame):
            self._frame = frame
        else:
            self._frame = Frame(frame)

    @property
    def frame(self):
        """
        The Decorator delegates all work to the wrapped component.
        """

        return self._frame

    def apply(self, input_image):
        
        return self._frame.apply(input_image)





import cv2 as cv
from py2vision.image_process.frame_decorator import FrameDecorator
class Rotate(FrameDecorator):
    """
    Concrete Decorators call the wrapped object and alter its result in some
    way.
    """
    def apply(self, angle):
        """Apply rotation on an image.

        Args: 
            angle: an integer that represents rotation angle.

        Returns: 
            an image rotated.
        """
        # dividing height and width by 2 to get the center of the image
        height, width = self._frame.apply().shape[:2]
        # get the center coordinates of the image to create the 2D rotation matrix
        center = (width/2, height/2)
        # using cv2.getRotationMatrix2D() to get the rotation matrix
        rotate_matrix = cv.getRotationMatrix2D(center=center, angle=angle, scale=1)
        return cv.warpAffine(src=self._frame.apply(), M=rotate_matrix, dsize=(width, height))


from py2vision.image_process.frame_decorator import FrameDecorator

class SplitPair(FrameDecorator):
    """
    Concrete Decorators call the wrapped object and alter its result in some
    way.
    """
    def apply(self, mode="sbs"):
        """Apply split in an image pair from stereo cameras.

        Args:
            mode: if it is "sbs" (side-by-side) it will do a vertical slice, if it is "tb" (top-bottom) it will do a a horizontal slice.
            
        Returns: 
            left image and right image.
        """
        # dividing height and width by 2
        height, width = self._frame.apply().shape[:2]
        new_width = int(width/2)
        new_height = int(height/2)
        if mode == "sbs":
            img_left = self._frame.apply()[0:height,0:new_width] #Y+H and X+W
            img_right = self._frame.apply()[0:height,new_width:width]
        elif mode == "tb":
            img_left = self._frame.apply()[0:new_height, 0:width] #Y+H and X+W
            img_right = self._frame.apply()[new_height:height, 0:width]
        else:
            ValueError("mode just can be 'sbs' or 'tb'")
        return img_left, img_right


import tensorflow as tf

from py2vision.models.layers.conv2d_bn_leaky_relu_layer import conv2d_bn_leaky_relu_layer

def residual_layer(input_layer, input_channel, filter_num1, filter_num2):
    """ A residual block with two Convolutional-batch normalization-leakyReLu layers stacked one above other.

    Args: 
        input_layer: A tensor that works like input.
        input_channel: input layer dimensions.
        filter_num1: filter depth for the first convolutional-batch normalization-leakyRelu layer.
        filter_num2: filter depth for the second convolutional-batch normalization-leakyRelu layer.

    Returns:
        A residual block.
    """
    short_cut = input_layer
    x = conv2d_bn_leaky_relu_layer(input_layer, filters_shape=(1, 1, input_channel, filter_num1))
    x = conv2d_bn_leaky_relu_layer(x, filters_shape=(3, 3, filter_num1,   filter_num2))
    residual_output = short_cut + x

    return residual_output


import tensorflow as tf

from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import LeakyReLU
from tensorflow.keras.layers import ZeroPadding2D
from tensorflow.keras.regularizers import l2

from py2vision.models.layers.batch_normalization_layer import BatchNormalization

def conv2d_bn_leaky_relu_layer(input_layer, filters_shape, downsample=False, activate=True, bn=True):
    """ A resnet block with depthwise separable convolutions to reduce the computational demand.

    Args: 
        input_layer: A tensor that works like input.
        filters_shape: An array or list or tuple that contains the shape of the filters which filters_shape[0] is kernel size for conv2d layer.
        downsample: a boolean to know when apply zero padding layer.
        activate: a boolean to know when apply leaky ReLu layer.
        bn: a boolean to know when apply a batch normalization layer.

    Returns: 
        A resnet block
    """
    if downsample:
        input_layer = ZeroPadding2D(((1, 0), (1, 0)))(input_layer)
        padding = 'valid'
        strides = 2
    else:
        strides = 1
        padding = 'same'

    x = Conv2D(filters=filters_shape[-1], kernel_size = filters_shape[0], strides=strides,
                  padding=padding, use_bias=not bn, kernel_regularizer=l2(0.0005),
                  kernel_initializer=tf.random_normal_initializer(stddev=0.01),
                  bias_initializer=tf.constant_initializer(0.))(input_layer)
    if bn:
        x = BatchNormalization()(x)
    if activate == True:
        x = LeakyReLU(alpha=0.1)(x)

    return x


import numpy as np
import tensorflow as tf

class YoloV3Calculus:
    """
    Useful methods for calculating the IOU, decode 
    network output when training, nms, yolov3 loss, and bounding box offsets.
    and bounding box offsets.
    """

    def decode(self, conv_output, num_class, i=0, strides=[8, 16, 32], 
                anchors=[[[10,  13], [16,   30], [33,   23]],
                        [[30,  61], [62,   45], [59,  119]],
                        [[116, 90], [156, 198], [373, 326]]]):
        """
        A piece of code that receives convolutional layers and returns the prediction layers.
        
        Args:
            conv_output: output of the Yolo model.
            num_class: an integer representing how many classes the model has.
            i: an integer that can be 0, 1 or 2 to correspond to the three scales of the grid.
            strides: a list with a length of 3 corresponding to the strides of the prediction layer.
            anchors: a 3-dimensional list of anchor sizes.
        
        Returns:
            the predicted probability category box object.
        """
        strides = np.array(strides)
        anchors = (np.array(anchors).T/strides).T
        # where i = 0, 1 or 2 to correspond to the three grid scales  
        conv_shape       = tf.shape(conv_output)
        batch_size       = conv_shape[0]
        output_size      = conv_shape[1]

        conv_output = tf.reshape(conv_output, (batch_size, output_size, output_size, 3, 5 + num_class))

        conv_raw_dxdy = conv_output[:, :, :, :, 0:2] # offset of center position     
        conv_raw_dwdh = conv_output[:, :, :, :, 2:4] # Prediction box length and width offset
        conv_raw_conf = conv_output[:, :, :, :, 4:5] # confidence of the prediction box
        conv_raw_prob = conv_output[:, :, :, :, 5: ] # category probability of the prediction box 

        # next need Draw the grid. Where output_size is equal to 13, 26 or 52  
        y = tf.range(output_size, dtype=tf.int32)
        y = tf.expand_dims(y, -1)
        y = tf.tile(y, [1, output_size])
        x = tf.range(output_size,dtype=tf.int32)
        x = tf.expand_dims(x, 0)
        x = tf.tile(x, [output_size, 1])

        xy_grid = tf.concat([x[:, :, tf.newaxis], y[:, :, tf.newaxis]], axis=-1)
        xy_grid = tf.tile(xy_grid[tf.newaxis, :, :, tf.newaxis, :], [batch_size, 1, 1, 3, 1])
        xy_grid = tf.cast(xy_grid, tf.float32)

        # Calculate the center position of the prediction box:
        pred_xy = (tf.sigmoid(conv_raw_dxdy) + xy_grid) * strides[i]
        # Calculate the length and width of the prediction box:
        pred_wh = (tf.exp(conv_raw_dwdh) * anchors[i]) * strides[i]

        pred_xywh = tf.concat([pred_xy, pred_wh], axis=-1)
        pred_conf = tf.sigmoid(conv_raw_conf) # object box calculates the predicted confidence
        pred_prob = tf.sigmoid(conv_raw_prob) # calculating the predicted probability category box object

        # calculating the predicted probability category box object
        return tf.concat([pred_xywh, pred_conf, pred_prob], axis=-1)

    def centroid2minmax(self, boxes):
        """Centroid to minmax format (cx, cy, w, h) to (xmin, ymin, xmax, ymax).

        Args:
            boxes: Batch of  bounding boxes in centroid format.

        Returns:
            minmax: Batch of boxes in minmax format
        """
        minmax= np.copy(boxes).astype(np.float)
        minmax[..., 0] = boxes[..., 0] - (0.5 * boxes[..., 2])
        minmax[..., 1] = boxes[..., 1] - (0.5 * boxes[..., 3])
        minmax[..., 2] = boxes[..., 0] + (0.5 * boxes[..., 2])
        minmax[..., 3] = boxes[..., 1] + (0.5 * boxes[..., 3])
        return minmax

    def minmax2centroid(self, boxes):
        """Minmax to centroid format (xmin, ymin, xmax, ymax) to (cx, cy, w, h).
        
        Arguments:
            boxes: Batch of bounding boxes in minmax format.

        Returns:
            A Batch of boxes in centroid format
        """
        centroid = np.copy(boxes).astype(np.float)
        centroid[..., 0] = 0.5 * (boxes[..., 2] - boxes[..., 0])
        centroid[..., 0] += boxes[..., 0] 
        centroid[..., 1] = 0.5 * (boxes[..., 3] - boxes[..., 1])
        centroid[..., 1] += boxes[..., 1] 
        centroid[..., 2] = boxes[..., 2] - boxes[..., 0]
        centroid[..., 3] = boxes[..., 3] - boxes[..., 1]
        return centroid
    
    def bbox_iou(self, boxes1, boxes2):
        """Compute Intersection Over Union between anchor boxes and bounding boxes.
        
        Args:
            boxes1: an array or tensor with a shape (n, 4).
            boxes2: an array or tensor with a shape (n, 4).
        
        Returns:
            A value between (0, 1) that correspond with IoU.
        """
        boxes1_area = boxes1[..., 2] * boxes1[..., 3]
        boxes2_area = boxes2[..., 2] * boxes2[..., 3]

        boxes1 = tf.concat([boxes1[..., :2] - boxes1[..., 2:] * 0.5,
                            boxes1[..., :2] + boxes1[..., 2:] * 0.5], axis=-1)
        boxes2 = tf.concat([boxes2[..., :2] - boxes2[..., 2:] * 0.5,
                            boxes2[..., :2] + boxes2[..., 2:] * 0.5], axis=-1)

        left_up = tf.maximum(boxes1[..., :2], boxes2[..., :2])
        right_down = tf.minimum(boxes1[..., 2:], boxes2[..., 2:])

        inter_section = tf.maximum(right_down - left_up, 0.0)
        inter_area = inter_section[..., 0] * inter_section[..., 1]
        union_area = boxes1_area + boxes2_area - inter_area

        return 1.0 * inter_area / union_area

    def bbox_giou(self, boxes1, boxes2):
        """
        Compute Generalized Intersection Over Union between bounding boxes.
        
        Args:
            boxes1: an array or tensor with a shape (n, 4).
            boxes2: an array or tensor with a shape (n, 4).
        
        Returns:
            A value between (0, 1) that correspond with GIoU.
        """
        boxes1 = tf.concat([boxes1[..., :2] - boxes1[..., 2:] * 0.5,
                            boxes1[..., :2] + boxes1[..., 2:] * 0.5], axis=-1)
        boxes2 = tf.concat([boxes2[..., :2] - boxes2[..., 2:] * 0.5,
                            boxes2[..., :2] + boxes2[..., 2:] * 0.5], axis=-1)

        boxes1 = tf.concat([tf.minimum(boxes1[..., :2], boxes1[..., 2:]),
                            tf.maximum(boxes1[..., :2], boxes1[..., 2:])], axis=-1)
        boxes2 = tf.concat([tf.minimum(boxes2[..., :2], boxes2[..., 2:]),
                            tf.maximum(boxes2[..., :2], boxes2[..., 2:])], axis=-1)

        boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
        boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])

        left_up = tf.maximum(boxes1[..., :2], boxes2[..., :2])
        right_down = tf.minimum(boxes1[..., 2:], boxes2[..., 2:])

        inter_section = tf.maximum(right_down - left_up, 0.0)
        inter_area = inter_section[..., 0] * inter_section[..., 1]
        union_area = boxes1_area + boxes2_area - inter_area

        # Calculate the iou value between the two bounding boxes
        iou = inter_area / union_area

        # Calculate the coordinates of the upper left corner and the lower right corner of the smallest closed convex surface
        enclose_left_up = tf.minimum(boxes1[..., :2], boxes2[..., :2])
        enclose_right_down = tf.maximum(boxes1[..., 2:], boxes2[..., 2:])
        enclose = tf.maximum(enclose_right_down - enclose_left_up, 0.0)

        # Calculate the area of the smallest closed convex surface C
        enclose_area = enclose[..., 0] * enclose[..., 1]

        # Calculate the GIoU value according to the GioU formula  
        giou = iou - 1.0 * (enclose_area - union_area) / enclose_area

        return giou

    def bbox_ciou(self, boxes1, boxes2):
        """Compute Complete Intersection Over Union between bounding boxes.
        
        Args:
            boxes1: an array or tensor with a shape (n, 4).
            boxes2: an array or tensor with a shape (n, 4).
        
        Returns:
            A value between (0, 1) that correspond with CIoU.
        """
        boxes1_coor = tf.concat([boxes1[..., :2] - boxes1[..., 2:] * 0.5,
                            boxes1[..., :2] + boxes1[..., 2:] * 0.5], axis=-1)
        boxes2_coor = tf.concat([boxes2[..., :2] - boxes2[..., 2:] * 0.5,
                            boxes2[..., :2] + boxes2[..., 2:] * 0.5], axis=-1)

        left = tf.maximum(boxes1_coor[..., 0], boxes2_coor[..., 0])
        up = tf.maximum(boxes1_coor[..., 1], boxes2_coor[..., 1])
        right = tf.maximum(boxes1_coor[..., 2], boxes2_coor[..., 2])
        down = tf.maximum(boxes1_coor[..., 3], boxes2_coor[..., 3])

        c = (right - left) * (right - left) + (up - down) * (up - down)
        iou = self.bbox_iou(boxes1, boxes2)

        u = (boxes1[..., 0] - boxes2[..., 0]) * (boxes1[..., 0] - boxes2[..., 0]) + (boxes1[..., 1] - boxes2[..., 1]) * (boxes1[..., 1] - boxes2[..., 1])
        d = u / c

        ar_gt = boxes2[..., 2] / boxes2[..., 3]
        ar_pred = boxes1[..., 2] / boxes1[..., 3]

        ar_loss = 4 / (np.pi * np.pi) * (tf.atan(ar_gt) - tf.atan(ar_pred)) * (tf.atan(ar_gt) - tf.atan(ar_pred))
        alpha = ar_loss / (1 - iou + ar_loss + 0.000001)
        ciou_term = d + alpha * ar_loss

        return iou - ciou_term
    
    def loss(self, pred, conv, label, bboxes, num_class, i=0, strides=[8, 16, 32], loss_thresh=0.5):
        """Calculate a loss vector to train a yolo network using GIoU, confidence and probability losses.
        
        Args:
            pred: the prediction of the model.
            conv: the last convolutional layer of a yolo model.
            label: expected label.
            bboxes: ground truth.
            num_class: an integer with the number of classes to detect.
            i: an integer which can be 0, 1, or 2 to correspond to the three grid scales.
            strides: a list with a len of 3 that correspond with the strides between each prediction.
            loss_thresh: a number between (0, 1) which if IoU is less than it, it is considered that the prediction box contains no objects.
        
        Returns:
            A tuple with a len of 3 where the first argument is the GIoU loss, next
            Confidence loss and the last one the probability loss.
        """
        strides = np.array(strides)

        conv_shape  = tf.shape(conv)
        batch_size  = conv_shape[0]
        output_size = conv_shape[1]
        input_size  = strides[i] * output_size
        conv = tf.reshape(conv, (batch_size, output_size, output_size, 3, 5 + num_class))

        conv_raw_conf = conv[:, :, :, :, 4:5]
        conv_raw_prob = conv[:, :, :, :, 5:]

        pred_xywh     = pred[:, :, :, :, 0:4]
        pred_conf     = pred[:, :, :, :, 4:5]

        label_xywh    = label[:, :, :, :, 0:4]
        respond_bbox  = label[:, :, :, :, 4:5]
        label_prob    = label[:, :, :, :, 5:]

        giou = tf.expand_dims(self.bbox_giou(pred_xywh, label_xywh), axis=-1)
        input_size = tf.cast(input_size, tf.float32)

        bbox_loss_scale = 2.0 - 1.0 * label_xywh[:, :, :, :, 2:3] * label_xywh[:, :, :, :, 3:4] / (input_size ** 2)
        giou_loss = respond_bbox * bbox_loss_scale * (1 - giou)

        iou = self.bbox_iou(pred_xywh[:, :, :, :, np.newaxis, :], bboxes[:, np.newaxis, np.newaxis, np.newaxis, :, :])
        # Find the value of IoU with the real box The largest prediction box
        max_iou = tf.expand_dims(tf.reduce_max(iou, axis=-1), axis=-1)

        # If the largest iou is less than the threshold, it is considered that the prediction box contains no objects, then the background box
        respond_bgd = (1.0 - respond_bbox) * tf.cast( max_iou < loss_thresh, tf.float32 )

        conf_focal = tf.pow(respond_bbox - pred_conf, 2)

        # Calculate the loss of confidence
        # we hope that if the grid contains objects, then the network output prediction box has a confidence of 1 and 0 when there is no object.
        conf_loss = conf_focal * (
                respond_bbox * tf.nn.sigmoid_cross_entropy_with_logits(labels=respond_bbox, logits=conv_raw_conf)
                +
                respond_bgd * tf.nn.sigmoid_cross_entropy_with_logits(labels=respond_bbox, logits=conv_raw_conf)
        )

        prob_loss = respond_bbox * tf.nn.sigmoid_cross_entropy_with_logits(labels=label_prob, logits=conv_raw_prob)

        giou_loss = tf.reduce_mean(tf.reduce_sum(giou_loss, axis=[1,2,3,4]))
        conf_loss = tf.reduce_mean(tf.reduce_sum(conf_loss, axis=[1,2,3,4]))
        prob_loss = tf.reduce_mean(tf.reduce_sum(prob_loss, axis=[1,2,3,4]))

        return giou_loss, conf_loss, prob_loss
    
    def best_bboxes_iou(self, boxes1, boxes2):
        """
        Compute Intersection Over Union between bounding boxes and return the best choices to apply nms algorithm.
        
        Args:
            boxes1: an array or tensor with a shape (n, 4).
            boxes2: an array or tensor with a shape (n, 4).
        
        Returns:
            A value between (0, 1) that correspond with IoU.
        """
        boxes1 = np.array(boxes1)
        boxes2 = np.array(boxes2)

        boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
        boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])

        left_up       = np.maximum(boxes1[..., :2], boxes2[..., :2])
        right_down    = np.minimum(boxes1[..., 2:], boxes2[..., 2:])

        inter_section = np.maximum(right_down - left_up, 0.0)
        inter_area    = inter_section[..., 0] * inter_section[..., 1]
        union_area    = boxes1_area + boxes2_area - inter_area
        ious          = np.maximum(1.0 * inter_area / union_area, np.finfo(np.float32).eps)

        return ious

    def nms(self, bboxes, iou_threshold, sigma=0.3, method='nms'):
        """
        Compute Non maximum supression algorithm.
        Note: see this paper to understand soft-nms https://arxiv.org/pdf/1704.04503.pdf.

        Args:
            bboxes: (xmin, ymin, xmax, ymax, score, class).
            iou_threshold: a parameter between (0, 1).
            sigma: a parameter between (0, 1).
            method: a string that can be  'nms' or 'soft-nms'.

        Returns:
            Better bounding boxes.
        """
        classes_in_img = list(set(bboxes[:, 5]))
        best_bboxes = []

        for cls in classes_in_img:
            cls_mask = (bboxes[:, 5] == cls)
            cls_bboxes = bboxes[cls_mask]
            # Process 1: Determine whether the number of bounding boxes is greater than 0 
            while len(cls_bboxes) > 0:
                # Process 2: Select the bounding box with the highest score according to socre order A
                max_ind = np.argmax(cls_bboxes[:, 4])
                best_bbox = cls_bboxes[max_ind]
                best_bboxes.append(best_bbox)
                cls_bboxes = np.concatenate([cls_bboxes[: max_ind], cls_bboxes[max_ind + 1:]])
                # Process 3: Calculate this bounding box A and
                # Remain all iou of the bounding box and remove those bounding boxes whose iou value is higher than the threshold 
                iou = self.best_bboxes_iou(best_bbox[np.newaxis, :4], cls_bboxes[:, :4])
                weight = np.ones((len(iou),), dtype=np.float32)

                assert method in ['nms', 'soft-nms']

                if method == 'nms':
                    iou_mask = iou > iou_threshold
                    weight[iou_mask] = 0.0

                if method == 'soft-nms':
                    weight = np.exp(-(1.0 * iou ** 2 / sigma))

                cls_bboxes[:, 4] = cls_bboxes[:, 4] * weight
                score_mask = cls_bboxes[:, 4] > 0.
                cls_bboxes = cls_bboxes[score_mask]

        return best_bboxes

    def postprocess_boxes(self, pred_bbox, original_image, input_size, score_threshold):
        """
        Improve predicted bounding boxes and resize them.

        Args:
            pred_bbox: a predicted bonding box.
            original_image: an image before resizing.
            input_size: the dimension of original image after resizing like an square image.
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.

        Returns:
            Bounding boxes that are inside the range, valids and with a score greather than score_threshold.
        """
        valid_scale=[0, np.inf]
        pred_bbox = np.array(pred_bbox)

        pred_xywh = pred_bbox[:, 0:4]
        pred_conf = pred_bbox[:, 4]
        pred_prob = pred_bbox[:, 5:]

        # 1. (x, y, w, h) --> (xmin, ymin, xmax, ymax)
        pred_coor = np.concatenate([pred_xywh[:, :2] - pred_xywh[:, 2:] * 0.5,
                                    pred_xywh[:, :2] + pred_xywh[:, 2:] * 0.5], axis=-1)
        # 2. (xmin, ymin, xmax, ymax) -> (xmin_org, ymin_org, xmax_org, ymax_org)
        org_h, org_w = original_image.shape[:2]
        resize_ratio = min(input_size / org_w, input_size / org_h)

        dw = (input_size - resize_ratio * org_w) / 2
        dh = (input_size - resize_ratio * org_h) / 2

        pred_coor[:, 0::2] = 1.0 * (pred_coor[:, 0::2] - dw) / resize_ratio
        pred_coor[:, 1::2] = 1.0 * (pred_coor[:, 1::2] - dh) / resize_ratio

        # 3. clip some boxes those are out of range
        pred_coor = np.concatenate([np.maximum(pred_coor[:, :2], [0, 0]),
                                    np.minimum(pred_coor[:, 2:], [org_w - 1, org_h - 1])], axis=-1)
        invalid_mask = np.logical_or((pred_coor[:, 0] > pred_coor[:, 2]), (pred_coor[:, 1] > pred_coor[:, 3]))
        pred_coor[invalid_mask] = 0

        # 4. discard some invalid boxes
        bboxes_scale = np.sqrt(np.multiply.reduce(pred_coor[:, 2:4] - pred_coor[:, 0:2], axis=-1))
        scale_mask = np.logical_and((valid_scale[0] < bboxes_scale), (bboxes_scale < valid_scale[1]))

        # 5. discard boxes with low scores
        classes = np.argmax(pred_prob, axis=-1)
        scores = pred_conf * pred_prob[np.arange(len(pred_coor)), classes]
        score_mask = scores > score_threshold
        mask = np.logical_and(scale_mask, score_mask)
        coors, scores, classes = pred_coor[mask], scores[mask], classes[mask]

        return np.concatenate([coors, scores[:, np.newaxis], classes[:, np.newaxis]], axis=-1)


from __future__ import annotations
from abc import ABC, abstractmethod

import re
import wget
import os



class Recognizer:
    """
    An Abstraction that expects an implementation of a detector like:
    'ObjectDetectorYoloV3'.
    """

    def __init__(self, neural_network: NeuralNetwork):
        self.implementation = neural_network

    def get_model(self):
        """Returns a tensorflow model with an implemented network"""
        return self.implementation.model
    
    def print_model(self):
        """Print network summary for debugging purposes."""
        self.implementation.print_summary()

    def inference(self, image_path, input_size=416, score_threshold=0.3, iou_threshold=0.45, nms_method="nms"):
        """ Apply inference with trained model.

        Args:
            image_path: a path to an image.
            input_size: integer to resize an input image from their original dimensions to an square image.
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            nms_method: a string that can be  'nms' or 'soft-nms'.

        Returns:
            An array with bounding boxes.
        """
        return self.implementation.inference(image_path, input_size, score_threshold, iou_threshold, nms_method)

    def restore_weights(self, weights_file, use_checkpoint=False):
        """Load previously trained model weights.

        Args: 
            weights_file: beginning by project root this is the path where is save your weights; example: "weights/weights_01.h5".
            use_checkpoint: if you wanna use a .ckpt file this variable should be True.
        """
        self.implementation.restore_weights(weights_file, use_checkpoint)

    def train_using_weights(self, train_annotations_path, test_annotations_path, class_file_name, weights_path,
                checkpoint_path="checkpoints", use_checkpoint=False, warmup_epochs=2, 
                epochs=100, log_dir="logs", save_only_best_model=True, save_all_checkpoints=False, batch_size=4, lr_init=1e-4, lr_end=1e-6,
                strides=[8, 16, 32],
                anchors=[[[10,  13], [16,   30], [33,   23]],
                        [[30,  61], [62,   45], [59,  119]],
                        [[116, 90], [156, 198], [373, 326]]],
                anchor_per_scale=3, max_bbox_per_scale=100):
        """Train with transfer learning.
        
        Args:
            train_annotations_path: a string corresponding to the folder where train annotations are located.
            test_annotations_path: a string corresponding to the folder where test annotations are located.
            class_file_name: a string corresponding to the classes file (a .txt file with a list of classes) is located.
            weights_path: a path which in case if it's an url with weights like: 'https://pjreddie.com/media/files/yolov3.weights' or 'https://pjreddie.com/media/files/yolov3-tiny.weights' this method first download the weights and then train the net but if the path is a local file the method is going to load the weights and next train the network.
            checkpoint_path: a string corresponding to the checkpoint file that is inside of a checkpoints folder.
            use_checkpoint: a boolean that controls if use chepoint before train.
            warmup_epochs: an hiperparameter that update learning rate like this paper https://arxiv.org/pdf/1812.01187.pdf&usg=ALkJrhglKOPDjNt6SHGbphTHyMcT0cuMJg.
            epochs: Number of epochs to train.
            log_dir: a folder to save logs.
            save_only_best_model: if is true the model will be saved when best validation loss > total validation loss/total test elements, but if it isn't true model will be saved always.
            save_all_checkpoints: it is a boolean, if is true model will be saved in each epoch.
            batch_size: an integer with the size of batches in test and train datasets.
            lr_init: a float which is initial learning rate.
            lr_end: a float which is final learning rate.
            strides: a list with the strides in a yolo model.
            anchors: these are the yolo anchors sizes.
            anchor_per_scale: an integer with the number of anchor boxes per scale. 
            max_bbox_per_scale: nan integer with the number of bounding boxes per scale. 
        """
        regex = '((http|https)://)(www.)?' + '[a-zA-Z0-9@:%._\\+~#?&//=]{2,256}\\.[a-z]' + '{2,6}\\b([-a-zA-Z0-9@:%._\\+~#?&//=]*)'
        regular_exp = re.compile(regex)
        if re.search(regular_exp, weights_path):
            wget.download(weights_path)
            weights_path = os.path.basename(weights_path)
        # restore weights
        self.implementation.restore_weights(weights_path)
        # train with a pre-trained net
        self.implementation.train(train_annotations_path, test_annotations_path, class_file_name, checkpoint_path, use_checkpoint, warmup_epochs, epochs, log_dir, save_only_best_model, save_all_checkpoints, batch_size, lr_init, lr_end, strides=strides, anchors=anchors, anchor_per_scale=anchor_per_scale, max_bbox_per_scale=max_bbox_per_scale)

    def train(self, train_annotations_path, test_annotations_path, class_file_name, 
                checkpoint_path="checkpoints", use_checkpoint=False, warmup_epochs=2, 
                epochs=100, log_dir="logs", save_only_best_model=True, save_all_checkpoints=False, batch_size=4, lr_init=1e-4, lr_end=1e-6,
                strides=[8, 16, 32],
                anchors=[[[10,  13], [16,   30], [33,   23]],
                        [[30,  61], [62,   45], [59,  119]],
                        [[116, 90], [156, 198], [373, 326]]],
                anchor_per_scale=3, max_bbox_per_scale=100):
        """Train an ssd network.

        Args:
            train_annotations_path: a string corresponding to the folder where train annotations are located.
            test_annotations_path: a string corresponding to the folder where test annotations are located.
            class_file_name: a string corresponding to the classes file (a .txt file with a list of classes) is located.
            checkpoint_path: a string corresponding to the checkpoint file that is inside of a checkpoints folder.
            use_checkpoint: a boolean that controls if use chepoint before train. 
            warmup_epochs: an hiperparameter that update learning rate like this paper https://arxiv.org/pdf/1812.01187.pdf&usg=ALkJrhglKOPDjNt6SHGbphTHyMcT0cuMJg.
            epochs: Number of epochs to train.
            log_dir: a folder to save logs.
            save_only_best_model: if is true the model will be saved when best validation loss > total validation loss/total test elements, but if it isn't true model will be saved always.
            save_all_checkpoints: it is a boolean, if is true model will be saved in each epoch.
            batch_size: an integer with the size of batches in test and train datasets.
            lr_init: a float which is initial learning rate.
            lr_end: a float which is final learning rate.
            strides: a list with the strides in a yolo model.
            anchors: these are the yolo anchors sizes.
            anchor_per_scale: an integer with the number of anchor boxes per scale. 
            max_bbox_per_scale: nan integer with the number of bounding boxes per scale. 
        """
        self.implementation.train(train_annotations_path, test_annotations_path, class_file_name, checkpoint_path, use_checkpoint, warmup_epochs, epochs, log_dir, save_only_best_model, save_all_checkpoints, batch_size, lr_init, lr_end, strides=strides, anchors=anchors, anchor_per_scale=anchor_per_scale, max_bbox_per_scale=max_bbox_per_scale)
    
    def evaluate(self, model, dataset, classes_file, score_threshold=0.05, iou_threshold=0.50, test_input_size=416):
        """Apply evaluation using mAP.

        Args:
            model: a tesorflow detection model.
            dataset: an YoloV3DatasetGenerator instance with test dataset.
            classes_file: a string corresponding to the classes file (a .txt file with a list of classes) is located.
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm.
            test_input_size: integer to resize an input image from their original dimensions to an square image.

        Returns:
            mAP score
        """
        return self.implementation.evaluate(model, dataset, classes_file, score_threshold, iou_threshold, test_input_size)



class NeuralNetwork(ABC):
    """
    The Implementation defines the interface for all implementation classes. It doesn't have to match the Abstraction's interface. In fact, the two interfaces can be entirely different. Typically the Implementation interface provides only primitive operations, while the Abstraction defines higher-level operations based on those primitives.
    """

    @abstractmethod
    def train(self):
        pass
    
    @abstractmethod
    def inference(self):
        pass
    
    @abstractmethod
    def restore_weights(self):
        pass
    
    @abstractmethod
    def evaluate(self):
        pass
    
    @abstractmethod
    def print_summary(self):
        pass
    

    






import cv2 as cv

from py2vision.image_process.frame_decorator import FrameDecorator

class Resize(FrameDecorator):
    def apply(self, width, height):
        """Apply resizing on an image.
        
        Args:
             width: new width to the image.
             height: new height to the image.
        
        Returns:
            an image resized.
        """
        return cv.resize(self._frame.apply(), (width, height), interpolation= cv.INTER_LINEAR_EXACT)


from abc import ABC, abstractmethod

import cv2 as cv
import numpy as np
import tensorflow as tf
import time
import requests
import os 

from multiprocessing import Process, Queue
from py2vision.image_process.frame_decorator import Frame
from py2vision.image_process.resize_with_bbox import ResizeWithBBox
from py2vision.compute.yolov3_calculus import YoloV3Calculus
from py2vision.utils.draw import draw_bbox
from py2vision.input_output.camera import Camera

class DetectionMode(ABC):
    """
    The Abstract Class which defines detect method that contains the skeleton of detection algorithm with different methods such as detection on images, detection on video, detection using multiprocessing and real time detection.
    """
    def predict(self, model, image_data):
        """Apply a prediction step.

        Args:
            model: expects a tensorflow model trained.
            image_data: expects an array or tensor with image data.

        Returns:
            predicted bounding boxes.
        """
        pred_bbox = model.predict(image_data)
        pred_bbox = [tf.reshape(x, (-1, tf.shape(x)[-1])) for x in pred_bbox]

        return tf.concat(pred_bbox, axis=0)
    
    def postprocess_boxes(self, original_image, pred_bbox, input_size=416, score_threshold=0.3, iou_threshold=0.45, nms_method="nms"):
        """Apply a postprocess and non maximum supression step and resize bboxes.

        Args:
            original_image: expects a tensorflow model trained.
            pred_bbox: a tensor of predicted bounding boxes.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            nms_method: a string that can be  'nms' or 'soft-nms'.

        Returns:
            Improved bounding boxes
        """
        compute = YoloV3Calculus()
        bboxes = compute.postprocess_boxes(pred_bbox, original_image, input_size, score_threshold)

        return compute.nms(bboxes, iou_threshold, method=nms_method)

    def pre_process(self, image, input_size):
        """Apply resize images step.

        Args:
            image: expects a an array.
            input_size: integer to resize an input image from their original dimensions to an square image.

        Returns:
            Resized images and bounding boxes.
        """
        frame = Frame(np.copy(image))
        image_data = ResizeWithBBox(frame).apply([input_size, input_size])

        return image_data[np.newaxis, ...].astype(np.float32)
    
    def draw(self, original_image, bboxes, class_file_name, rectangle_colors, homogeneous_points=None, text_colors=(255,255,0)):
        """Draw bounding boxes on images.

        Args:
            original_image: an array which correspond with an image
            bboxes: their bounding boxes.
            class_file_name: a path with a .txt file where the classes are saved.
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            homogeneous_points: an array with dimensions n x 4 where each row is like (X, Y, Z, W). However if is None it won't be drawed.

        Returns:
            An image with bounding boxes and homogeneous coordinates.
        """
        return draw_bbox(original_image, bboxes, class_file_name=class_file_name, 
                rectangle_colors=rectangle_colors, homogeneous_points=homogeneous_points, text_colors=text_colors)
    
    @abstractmethod
    def detect(self):
        """Apply detection pipeline"""
        pass
        
    @abstractmethod
    def prepare_input(self):
        """It's the first step to convert inputs like video paths or images in compatible data."""
        pass
    
    @abstractmethod
    def show(self):
        """Show actual frame in a window"""
        pass

    def camera_input(self, camera: Camera):
        """If camera source is webcam or other it is gonna create attribute 'vid', otherwise return a frame from streaming.
        
        Args:
            camera: is an instance of Camera Object
        """
        if camera.type_source == "other" or camera.type_source == "webcam":
            vid = cv.VideoCapture(camera.source)
            if not vid.isOpened():
                print("Cannot open camera")
                exit()
            return vid

        if camera.type_source == "stream":
            img_resp = requests.get(camera.source)
            img_arr = np.array(bytearray(img_resp.content), dtype=np.uint8)
            frame = cv.imdecode(img_arr, -1)
            return frame

class DetectRealTime(DetectionMode):
    def detect(self, model, camera: Camera, class_file_name, output_path="", input_size=416, 
                score_threshold=0.3, iou_threshold=0.45, rectangle_colors='', nms_method="nms", show=True):
        """Apply detection pipeline in realtime, if you want to close the window just press 'q'.
        
        Args:
            model: expects a tensorflow model trained.
            camera: An instance of camera object.
            class_file_name: it's the path of classes .txt file.
            output_path: if is an empty string, it won't be saved, but it is a path it save like a video.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm.
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.
            show: a boolean to show frame pcessed.
        """
        times = []
        vid = self.camera_input(camera)
        
        if isinstance(vid, cv.VideoCapture):
            out, vid, fps = self.prepare_input(vid, output_path)
        
        while True:
            if isinstance(vid, cv.VideoCapture):
                ret, img = vid.read()
                if not ret: break
            else: 
                img = self.camera_input(camera)
            try:
                original_frame = cv.cvtColor(img, cv.COLOR_BGR2RGB)
                original_frame = cv.cvtColor(original_frame, cv.COLOR_BGR2RGB)
            except:
                break
            image_data = self.pre_process(original_frame, input_size)
            t1 = time.time()
            pred_bbox = self.predict(model, image_data)
            t2 = time.time()
            bboxes = self.postprocess_boxes(original_frame, pred_bbox, input_size,
                                             score_threshold, iou_threshold, nms_method)
            times.append(t2-t1)
            times = times[-20:]
            
            ms = sum(times)/len(times)*1000
            fps = 1000 / ms
            
            print("Time: {:.2f}ms, {:.1f} FPS".format(ms, fps))
            frame = self.draw(original_frame, bboxes, class_file_name, rectangle_colors)
            cv.putText(frame, "Time: {:.1f}FPS".format(fps), (0, 30),
                          cv.FONT_HERSHEY_COMPLEX_SMALL, 1, (0, 0, 255), 2)
            if output_path != '': out.write(frame)
            if show: 
                self.show(frame, camera)
                if cv.waitKey(25) & 0xFF == ord("q"):
                    cv.destroyAllWindows()
                    break
        if show: cv.destroyAllWindows()

    def prepare_input(self, vid, output_path):
        """Initialization for writing and reading videos
        
        Args: 
            vid: a cv.VideoCapture instance.
            output_path: if is an empty string, it won't be saved, but it is a path it save like an image.

        Returns: 
            a tuple where its first argument is an instance to Opencv video writes, the second element is an instance to read frames, and the last element is the frames per second
        """
        if os.path.splitext(output_path)[1] != ".mp4":
                raise ValueError("output_path has to be an .mp4 file")
        # by default VideoCapture returns float instead of int
        width = int(vid.get(cv.CAP_PROP_FRAME_WIDTH))
        height = int(vid.get(cv.CAP_PROP_FRAME_HEIGHT))
        fps = int(vid.get(cv.CAP_PROP_FPS))
        codec = cv.VideoWriter_fourcc(*'XVID')
        out = cv.VideoWriter(output_path, codec, fps, (width, height)) # output_path must be .mp4

        return out, vid, fps 
    
    def show(self, image, camera: Camera):
        """Show an image

        Args: 
            image: an image array.
            camera: An instance of camera object.
        """
        cv.imshow(camera.id, image)

class DetectRealTimeMP(DetectionMode):
    def detect(self, model, camera: Camera, class_file_name, output_path="", input_size=416, 
                score_threshold=0.3, iou_threshold=0.45, rectangle_colors='', nms_method="nms"):
        """Apply detection pipeline using multiprocessing, if you want to close the window just press 'q'.
        
        Args:
            model: expects a tensorflow model trained.
            camera: An instance of camera object.
            class_file_name: it's the path of classes .txt file
            output_path: if is an empty string, it won't be saved, but it is a path it save like a video.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.
        """
        vid = self.camera_input(camera)

        if isinstance(vid, cv.VideoCapture):
            out, vid, fps = self.prepare_input(vid, output_path)
        original_frames = Queue()
        frames_data = Queue()
        predicted_data = Queue()
        processed_frames = Queue()
        processing_times = Queue()
        final_frames = Queue()
        p1, p2, p3 = self.multi_process_initialization(model, original_frames, frames_data, predicted_data, processed_frames, processing_times, final_frames, class_file_name, input_size, score_threshold, iou_threshold, rectangle_colors, nms_method)
        p1.start()
        p2.start()
        p3.start()
        while True:
            if isinstance(vid, cv.VideoCapture):
                ret, img = vid.read()
                if not ret: break
            else: 
                img = self.camera_input(camera)
            
            original_image = cv.cvtColor(img, cv.COLOR_BGR2RGB)
            original_image = cv.cvtColor(original_image, cv.COLOR_BGR2RGB)
            original_frames.put(original_image)
            image_data = self.pre_process(original_image, input_size)
            frames_data.put(image_data)
        
        while True:
            if original_frames.qsize() == 0 and frames_data.qsize() == 0 and predicted_data.qsize() == 0  and processed_frames.qsize() == 0  and processing_times.qsize() == 0 and final_frames.qsize() == 0:
                p1.terminate()
                p2.terminate()
                p3.terminate()
                break
            elif final_frames.qsize()>0:
                image = final_frames.get()
                if output_path != '': out.write(image)

        cv.destroyAllWindows()

    def prepare_input(self, vid, output_path) -> None:
        """Initialization for writing and reading videos.

        Args: 
            vid: cv.VideoCapture instance
            output_path: if is an empty string, it won't be saved, but it is a path it save like an image.

        Returns: 
            a tuple where its first argument is an instance to Opencv video writes, the second element is an instance to read frames,
            and the last element is the frames per second
        """
        if os.path.splitext(output_path)[1] != ".mp4":
            raise ValueError("output_path has to be an .mp4 file")
        # by default VideoCapture returns float instead of int
        width = int(vid.get(cv.CAP_PROP_FRAME_WIDTH))
        height = int(vid.get(cv.CAP_PROP_FRAME_HEIGHT))
        fps = int(vid.get(cv.CAP_PROP_FPS))
        codec = cv.VideoWriter_fourcc(*'XVID')
        out = cv.VideoWriter(output_path, codec, fps, (width, height)) # output_path must be .mp4
        no_of_frames = int(vid.get(cv.CAP_PROP_FRAME_COUNT))

        return out, vid, fps 

    def predict_bbox_mp(self, model, frames_data, predicted_data, processing_times):
        """ predict bounding boxes using multiprocessing. It needs to be initialized.

        Args: 
            model: expects a tensorflow model trained.
            frames_data: a queue from multiprocessing package, that corresponds with frames. 
            predicted_data: a queue from multiprocessing package, that corresponds with predictions.
            processing_times: a queue from multiprocessing package, that corresponds with times.
        """
        gpus = tf.config.experimental.list_physical_devices('GPU')
        if len(gpus) > 0:
            try: tf.config.experimental.set_memory_growth(gpus[0], True)
            except RuntimeError: print("RuntimeError in tf.config.experimental.list_physical_devices('GPU')")
        times = []
        while True:
            if frames_data.qsize()>0:
                image_data = frames_data.get()
                t1 = time.time()
                processing_times.put(time.time())
                pred_bbox = self.predict(model, image_data)
                
                predicted_data.put(pred_bbox)
    
    def postprocess_mp(self, predicted_data, original_frames, processed_frames, processing_times, input_size, class_file_name, score_threshold, iou_threshold, rectangle_colors, nms_method):
        """ Improve bounding boxes using multiprocessing. It needs to be initialized.

        Args: 
            predicted_data: a queue from multiprocessing package, that corresponds with predictions.
            original_frames: a queue from multiprocessing package, that corresponds with frames. 
            processed_frames: a queue from multiprocessing package, that corresponds with processed frames. 
            processing times: a queue from multiprocessing package, that corresponds with times.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            class_file_name: it's the path of classes .txt file
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.
        """
        times = []
        while True:
            if predicted_data.qsize()>0:
                pred_bbox = predicted_data.get()
                while original_frames.qsize() > 1:
                    original_image = original_frames.get()
                bboxes = self.postprocess_boxes(original_image, pred_bbox, input_size, score_threshold, iou_threshold, nms_method)
                image = self.draw(original_image, bboxes, class_file_name, rectangle_colors)
                times.append(time.time()-processing_times.get())
                times = times[-20:]
                
                ms = sum(times)/len(times)*1000
                fps = 1000 / ms
                image = cv.putText(image, "Time: {:.1f}FPS".format(fps), (0, 30), cv.FONT_HERSHEY_COMPLEX_SMALL, 1, (0, 0, 255), 2)
                #print("Time: {:.2f}ms, Final FPS: {:.1f}".format(ms, fps))
                
                processed_frames.put(image)

    def show(self, processed_frames, final_frames):
        """ Show preocessed input.
        
        Args: 
            processed_frames: a queue from multiprocessing package, that corresponds with processed frames. 
            final_frames: a queue from multiprocessing package, that corresponds with final frames. 
        """
        while True:
            if processed_frames.qsize()>0:
                image = processed_frames.get()
                final_frames.put(image)
                cv.imshow('Detection output', image)
                if cv.waitKey(25) & 0xFF == ord("q"):
                    cv.destroyAllWindows()
                    break

    def multi_process_initialization(self, model, original_frames, frames_data, predicted_data, processed_frames, processing_times, final_frames, class_file_name, input_size=416, score_threshold=0.3, iou_threshold=0.45, rectangle_colors='', nms_method="nms"):
        """Apply an initialite multi processing prediction, postprocessing and show steps.
        
        Args:
            model: expects a tensorflow model trained.
            original_frames: a queue from multiprocessing package, that corresponds with frames. 
            frames_data: a queue from multiprocessing package, that corresponds with frames. 
            predicted_data: a queue from multiprocessing package, that corresponds with predictions.
            processed_frames: a queue from multiprocessing package, that corresponds with processed frames. 
            processing times: a queue from multiprocessing package, that corresponds with times.
            final_frames: a queue from multiprocessing package, that corresponds with final frames. 
            class_file_name: it's the path of classes .txt file
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.

        Returns: 
            an Process tuple (prediction, postprocessing, show)
        """
        p1 = Process(target=self.predict_bbox_mp, args=(model, frames_data, predicted_data, processing_times))
        p2 = Process(target=self.postprocess_mp, args=(predicted_data, original_frames, processed_frames, processing_times, input_size, class_file_name, score_threshold, iou_threshold, rectangle_colors, nms_method))
        p3 = Process(target=self.show, args=(processed_frames, final_frames))
        return p1, p2, p3


class DetectVideo(DetectionMode):
    def detect(self, model, input_path, class_file_name, output_path="", input_size=416, 
                score_threshold=0.3, iou_threshold=0.45, rectangle_colors='', nms_method="nms", show=True):
        """Apply detection pipeline in a saved video, if you want to close the window just press 'q'.
        
        Args:
            model: expects a tensorflow model trained.
            input_path: a video path.
            class_file_name: it's the path of classes .txt file
            output_path: if is an empty string, it won't be saved, but it is a path it save like a video.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.
            show: a boolean to show frame pcessed
        """
        times, times_2 = [], []

        out, vid, fps = self.prepare_input(input_path, output_path)

        while True:
            _, img = vid.read()

            try:
                original_image = cv.cvtColor(img, cv.COLOR_BGR2RGB)
                original_image = cv.cvtColor(original_image, cv.COLOR_BGR2RGB)
            except:
                break
            
            image_data = self.pre_process(original_image, input_size)
            t1 = time.time()
            pred_bbox = self.predict(model, image_data)
            t2 = time.time()
            bboxes = self.postprocess_boxes(original_image, pred_bbox, input_size, score_threshold, iou_threshold, nms_method)
            image = self.draw(original_image, bboxes, class_file_name, rectangle_colors)
            t3 = time.time()
            times.append(t2-t1)
            times_2.append(t3-t1)
            
            times = times[-20:]
            times_2 = times_2[-20:]

            ms = sum(times)/len(times)*1000
            fps = 1000 / ms
            fps2 = 1000 / (sum(times_2)/len(times_2)*1000)
            
            image = cv.putText(image, "Time: {:.1f}FPS".format(fps), (0, 30), cv.FONT_HERSHEY_COMPLEX_SMALL, 1, (0, 0, 255), 2)
            # CreateXMLfile("XML_Detections", str(int(time.time())), original_image, bboxes, read_class_names(CLASSES))
            
            print("Time: {:.2f}ms, Detection FPS: {:.1f}, total FPS: {:.1f}".format(ms, fps, fps2))
            if output_path != '': out.write(image)
            if show: 
                self.show(image)
                if cv.waitKey(25) & 0xFF == ord("q"):
                    cv.destroyAllWindows()
                    break
        if show: cv.destroyAllWindows()

    def prepare_input(self, input_path, output_path):
        """Initialization for writing and reading videos.

        Args: 
            input_path: an video path.
            output_path: if is an empty string, it won't be saved, but it is a path it save like an image.
        
        Returns: 
            a tuple where its first argument is an instance to Opencv video writes, the second element is an instance to read frames,
            and the last element is the frames per second
        """
        if os.path.splitext(output_path)[1] != ".mp4":
            raise ValueError("output_path has to be an .mp4 file")
        vid = cv.VideoCapture(input_path)
        # by default VideoCapture returns float instead of int
        width = int(vid.get(cv.CAP_PROP_FRAME_WIDTH))
        height = int(vid.get(cv.CAP_PROP_FRAME_HEIGHT))
        fps = int(vid.get(cv.CAP_PROP_FPS))
        codec = cv.VideoWriter_fourcc(*'XVID')
        out = cv.VideoWriter(output_path, codec, fps, (width, height)) # output_path must be .mp4

        return out, vid, fps
    
    def show(self, image):
        """Show an image.
        Arguments: 
            image: an image array.
        """
        cv.imshow('Detection output', image)
        


class DetectImage(DetectionMode):
    def detect(self, model, input_path, class_file_name, output_path="", input_size=416,
                 score_threshold=0.3, iou_threshold=0.45, rectangle_colors='', nms_method="nms", show=True):
        """Apply detection pipeline in an image, if you want to close the window just press 'q'.
        
        Arguments:
            model: expects a tensorflow model trained.
            input_path: an image path
            class_file_name: it's the path of classes .txt file
            output_path: if is an empty string, it won't be saved, but it is a path it save like an image.
            input_size: integer to resize bounding boxes from their resized dimensions to original dimensions (input_size).
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            rectangle_colors: if this parameter is a string empty bounding box colors will be assing by default, however if rectangle_colors is a tuple like: (R, G, B) that will be bounding box colors.
            nms_method: a string that can be  'nms' or 'soft-nms'.
            show: a boolean to show frame pcessed.

        Returns:
            an image with prediction drawed
        """
        compatible_outputs = [".bmp", ".dib", ".jpg", ".jpeg", ".jpe", ".png", ".webp"]
        if not os.path.splitext(output_path)[1] in compatible_outputs: 
            raise ValueError("output_path only can be one of this: {}".format(compatible_outputs))
        original_image = self.prepare_input(input_path)
        image_data = self.pre_process(original_image, input_size)
        pred_bbox = self.predict(model, image_data)
        bboxes = self.postprocess_boxes(original_image, pred_bbox, input_size, score_threshold, iou_threshold, nms_method)
        image = self.draw(original_image, bboxes, class_file_name, rectangle_colors)
        if output_path != '': cv.imwrite(output_path, image)
        if show: self.show(image, output_path)

        return image

    def prepare_input(self, image_path):
        """Get a path an convert in an image array.

        Args: 
            image_path: a path.

        Returns: 
            An image array
        """
        original_image = cv.imread(image_path)
        original_image = cv.cvtColor(original_image, cv.COLOR_BGR2RGB)

        return  cv.cvtColor(original_image, cv.COLOR_BGR2RGB)
    
    def show(self, image, output_path=''):
        """Get a an image, then save it and finally show it.

        Args: 
            image: an image array.
            output_path: if is an empty string, it won't be saved, but it is a path it save like an image.
        """
        # Show the image
        cv.imshow("predicted image", image)
        # Load and hold the image
        cv.waitKey(0)
        # To close the window after the required kill value was provided
        cv.destroyAllWindows()

        



from abc import ABC, abstractmethod

from tensorflow.keras.layers import MaxPool2D

from py2vision.models.layers.conv2d_bn_leaky_relu_layer import conv2d_bn_leaky_relu_layer
from py2vision.models.layers.residual_layer import residual_layer

class BackboneStrategy(ABC):
    """
    The Strategy interface declares operations common to all supported versions
    of some algorithm.

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.
    """

    @abstractmethod
    def build(self):
        pass

class BackboneBlock():

    def __init__(self, strategy: BackboneStrategy):
        """
        Usually, the Context accepts a strategy through the constructor, but
        also provides a setter to change it at runtime.
        """

        self._strategy = strategy

    @property
    def strategy(self) -> BackboneStrategy:
        """
        The BackboneBlock maintains a reference to one of the Strategy objects. The
        BackboneBlock does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """

        return self._strategy
    
    @strategy.setter
    def strategy(self, strategy: BackboneStrategy):
        """
        To replacing a Strategy object at runtime.
        """

        self._strategy = strategy

    def build_model(self, input_shape):
        """Build a backbone for our net.
        
        Args:
            input_shape: Input image size and channels.

        Returns
            model (Keras Model)
        """
        if type(self._strategy).__name__ == 'darknet53':
            return self._strategy.build(input_shape)
        elif type(self._strategy).__name__ == 'darknet19_tiny':
            return self._strategy.build(input_shape)

class darknet53(BackboneStrategy):
    def build(self, x):
        """ Build a darknet53 model.

        Args: 
            x: an input tensor that can be an image.
            
        Returns:
            three branches of darknet53.
        """
        x = conv2d_bn_leaky_relu_layer(x, (3, 3,  3,  32))
        x = conv2d_bn_leaky_relu_layer(x, (3, 3, 32,  64), downsample=True)

        for i in range(1):
            x = residual_layer(x, 64,  32, 64)

        x = conv2d_bn_leaky_relu_layer(x, (3, 3,  64, 128), downsample=True)

        for i in range(2):
            x = residual_layer(x, 128,  64, 128)

        x = conv2d_bn_leaky_relu_layer(x, (3, 3, 128, 256), downsample=True)

        for i in range(8):
            x = residual_layer(x, 256, 128, 256)

        route_1 = x
        x = conv2d_bn_leaky_relu_layer(x, (3, 3, 256, 512), downsample=True)

        for i in range(8):
            x = residual_layer(x, 512, 256, 512)

        route_2 = x
        x = conv2d_bn_leaky_relu_layer(x, (3, 3, 512, 1024), downsample=True)

        for i in range(4):
            x = residual_layer(x, 1024, 512, 1024)

        return route_1, route_2, x

class darknet19_tiny(BackboneStrategy):
    def build(self, x):
        """ Build a darknet19 tiny model.
        
        Args: 
            x: an input tensor that can be an image.
        
        Returns: 
            Two branches of darknet19 tiny.
        """
        filters_shapes = [(3, 3, 3, 16), (3, 3, 16, 32), (3, 3, 32, 64), (3, 3, 64, 128), (3, 3, 128, 256)]
        for i, filter_shape in enumerate(filters_shapes):
            x = conv2d_bn_leaky_relu_layer(x, filter_shape)
            if i < 4:
                x = MaxPool2D(2, 2, 'same')(x)
        
        route_1 = x

        x = MaxPool2D(2, 2, 'same')(x)
        x = conv2d_bn_leaky_relu_layer(x, (3, 3, 256, 512))
        x = MaxPool2D(2, 1, 'same')(x)
        x = conv2d_bn_leaky_relu_layer(x, (3, 3, 512, 1024))

        return route_1, x
        


import tensorflow as tf

class UpsampleLayer(tf.keras.layers.Layer):
    """ A layer for upsampling an image. """
    def call(self, inputs):
        return tf.image.resize(inputs, (inputs.shape[1] * 2, inputs.shape[2] * 2), method='nearest')

        


import numpy as np
import cv2 as cv
import os
import random
import tensorflow as tf

from py2vision.utils.label_utils import read_class_names
from py2vision.image_process.frame_decorator import Frame
from py2vision.image_process.resize_with_bbox import ResizeWithBBox
from py2vision.compute.yolov3_calculus import YoloV3Calculus

class YoloV3DatasetGenerator(object):
    """
        A generator compatible with YOLO V3 network.
        
        Args:
            annotations_path: a path where the annotations are saved.
            class_file_name: a path with a .txt file where the classes are saved.
            batch_size: an integer that corresponds with the number of image which we introduce in a network per iteration.
            data_augmentation: a boolean that controls data augmentation which change original images in new ways.
            input_shape: a tuple with the input images dimensions.
            strides: a list with the strides in a yolo model.
            anchors: these are the yolo anchors sizes.
            anchor_per_scale: an integer with the number of anchor boxes per scale. 
            max_bbox_per_scale: nan integer with the number of bounding boxes per scale. 
            images_to_ram: a boolean to control when save images in ram which allow a faster training, but it needs more RAM. 
    """
    def __init__(self, annotations_path, class_file_name, batch_size=4, data_augmentation=True, 
                input_shape=(416, 416, 3), strides=[8, 16, 32],
                anchors=[[[10,  13], [16,   30], [33,   23]],
                        [[30,  61], [62,   45], [59,  119]],
                        [[116, 90], [156, 198], [373, 326]]],
                anchor_per_scale=3, max_bbox_per_scale=100, images_to_ram=True):
        self.annot_path  = annotations_path
        self.input_sizes = input_shape[0]
        self.batch_size  = batch_size
        self.data_aug    = data_augmentation

        #self.train_input_sizes = TRAIN_INPUT_SIZE
        self.strides = np.array(strides)
        self.classes = read_class_names(class_file_name)
        self.num_classes = len(self.classes)
        self.anchors = (np.array(anchors).T/self.strides).T
        self.anchor_per_scale = anchor_per_scale
        self.max_bbox_per_scale = max_bbox_per_scale
        self.images_to_ram = images_to_ram
        self.annotations = self.load_annotations()
        self.num_samples = len(self.annotations)
        self.num_batchs = int(np.ceil(self.num_samples / self.batch_size))
        self.batch_count = 0


    def load_annotations(self):
        """Returns annotations in array shape"""
        final_annotations = []
        with open(self.annot_path, 'r') as f:
            txt = f.readlines()
            annotations = [line.strip() for line in txt if len(line.strip().split()[1:]) != 0]
        np.random.shuffle(annotations)
        
        for annotation in annotations:
            # fully parse annotations
            line = annotation.split()
            image_path, index = "", 1
            for i, one_line in enumerate(line):
                if not one_line.replace(",","").isnumeric():
                    if image_path != "": image_path += " "
                    image_path += one_line
                else:
                    index = i
                    break
            if not os.path.exists(image_path):
                raise KeyError("%s does not exist ... " %image_path)
            if self.images_to_ram:
                image = cv.imread(image_path)
            else:
                image = ''
            final_annotations.append([image_path, line[index:], image])
        return final_annotations

    def __iter__(self):
        """Returns its own instance"""
        return self

    def delete_bad_annotation(self, bad_annotation):
        """Delete an annotation from annotations file .txt.

        Args:
            bad_annotation: a string with the path of an image
        """
        print(f'Deleting {bad_annotation} annotation line')
        bad_image_name = bad_annotation[0].split('/')[-1] # can be used to delete bad image

        # remove bad annotation line from annotation file
        with open(self.annot_path, "r+") as f:
            d = f.readlines()
            f.seek(0)
            for i in d:
                if bad_image_name not in i:
                    f.write(i)
            f.truncate()
    
    def __next__(self):
        """This method will be executed when you iterate this class

        Returns: 
            a batch of images when their bounding boxes
        """
        with tf.device('/cpu:0'):
            self.train_input_size = random.choice([self.input_sizes])
            self.train_output_sizes = self.train_input_size // self.strides

            batch_image = np.zeros((self.batch_size, self.train_input_size, self.train_input_size, 3), dtype=np.float32)

            batch_label_sbbox = np.zeros((self.batch_size, self.train_output_sizes[0], self.train_output_sizes[0],
                                          self.anchor_per_scale, 5 + self.num_classes), dtype=np.float32)
            batch_label_mbbox = np.zeros((self.batch_size, self.train_output_sizes[1], self.train_output_sizes[1],
                                          self.anchor_per_scale, 5 + self.num_classes), dtype=np.float32)
            batch_label_lbbox = np.zeros((self.batch_size, self.train_output_sizes[2], self.train_output_sizes[2],
                                          self.anchor_per_scale, 5 + self.num_classes), dtype=np.float32)

            batch_sbboxes = np.zeros((self.batch_size, self.max_bbox_per_scale, 4), dtype=np.float32)
            batch_mbboxes = np.zeros((self.batch_size, self.max_bbox_per_scale, 4), dtype=np.float32)
            batch_lbboxes = np.zeros((self.batch_size, self.max_bbox_per_scale, 4), dtype=np.float32)

            exceptions = False
            num = 0
            if self.batch_count < self.num_batchs:
                while num < self.batch_size:
                    index = self.batch_count * self.batch_size + num
                    if index >= self.num_samples: index -= self.num_samples
                    annotation = self.annotations[index]
                    image, bboxes = self.parse_annotation(annotation)
                    try:
                        label_sbbox, label_mbbox, label_lbbox, sbboxes, mbboxes, lbboxes = self.preprocess_true_boxes(bboxes)
                    except IndexError:
                        exceptions = True
                        self.delete_bad_annotation(annotation)
                        print("IndexError, something wrong with", annotation[0], "removed this line from annotation file")

                    batch_image[num, :, :, :] = image
                    batch_label_sbbox[num, :, :, :, :] = label_sbbox
                    batch_label_mbbox[num, :, :, :, :] = label_mbbox
                    batch_label_lbbox[num, :, :, :, :] = label_lbbox
                    batch_sbboxes[num, :, :] = sbboxes
                    batch_mbboxes[num, :, :] = mbboxes
                    batch_lbboxes[num, :, :] = lbboxes
                    num += 1

                if exceptions: 
                    print('\n')
                    raise Exception("There were problems with dataset, I fixed them, now restart the training process.")
                self.batch_count += 1
                batch_smaller_target = batch_label_sbbox, batch_sbboxes
                batch_medium_target  = batch_label_mbbox, batch_mbboxes
                batch_larger_target  = batch_label_lbbox, batch_lbboxes

                return batch_image, (batch_smaller_target, batch_medium_target, batch_larger_target)
            else:
                self.batch_count = 0
                np.random.shuffle(self.annotations)
                raise StopIteration

    def random_horizontal_flip(self, image, bboxes):
        """There is a probability of 0.5 to make an horizontal flip

        Args:  
            image: it can be an image or a batch of images.
            bboxes: this are corresponding bounding boxes.

        Returns
            images and their bounding boxes.
        """
        if random.random() < 0.5:
            _, w, _ = image.shape
            image = image[:, ::-1, :]
            bboxes[:, [0,2]] = w - bboxes[:, [2,0]]

        return image, bboxes

    def random_crop(self, image, bboxes):
        """There is a probability of 0.5 to make a crop.

        Args:  
            image: it can be an image or a batch of images.
            bboxes: this are corresponding bounding boxes.

        Returns
            images and their bounding boxes.
        """
        if random.random() < 0.5:
            h, w, _ = image.shape
            max_bbox = np.concatenate([np.min(bboxes[:, 0:2], axis=0), np.max(bboxes[:, 2:4], axis=0)], axis=-1)

            max_l_trans = max_bbox[0]
            max_u_trans = max_bbox[1]
            max_r_trans = w - max_bbox[2]
            max_d_trans = h - max_bbox[3]

            crop_xmin = max(0, int(max_bbox[0] - random.uniform(0, max_l_trans)))
            crop_ymin = max(0, int(max_bbox[1] - random.uniform(0, max_u_trans)))
            crop_xmax = max(w, int(max_bbox[2] + random.uniform(0, max_r_trans)))
            crop_ymax = max(h, int(max_bbox[3] + random.uniform(0, max_d_trans)))

            image = image[crop_ymin : crop_ymax, crop_xmin : crop_xmax]

            bboxes[:, [0, 2]] = bboxes[:, [0, 2]] - crop_xmin
            bboxes[:, [1, 3]] = bboxes[:, [1, 3]] - crop_ymin

        return image, bboxes

    def random_translate(self, image, bboxes):
        """There is a probability of 0.5 to shift in x and y images.

        Args:  
            image: it can be an image or a batch of images.
            bboxes: this are corresponding bounding boxes.

        Returns
            images and their bounding boxes.
        """
        if random.random() < 0.5:
            h, w, _ = image.shape
            max_bbox = np.concatenate([np.min(bboxes[:, 0:2], axis=0), np.max(bboxes[:, 2:4], axis=0)], axis=-1)

            max_l_trans = max_bbox[0]
            max_u_trans = max_bbox[1]
            max_r_trans = w - max_bbox[2]
            max_d_trans = h - max_bbox[3]

            tx = random.uniform(-(max_l_trans - 1), (max_r_trans - 1))
            ty = random.uniform(-(max_u_trans - 1), (max_d_trans - 1))

            M = np.array([[1, 0, tx], [0, 1, ty]])
            image = cv.warpAffine(image, M, (w, h))

            bboxes[:, [0, 2]] = bboxes[:, [0, 2]] + tx
            bboxes[:, [1, 3]] = bboxes[:, [1, 3]] + ty

        return image, bboxes

    def parse_annotation(self, annotation, mAP=False):
        """Convert an annotation in an image with their bounding boxes.

        Args:  
            annotation: an annotation line (path bounding box 1, class bounding box 2, class, ...)
            mAP: when it's true this method won't resize images and bounding boxes, otherwise it will do.

        Returns
            images and their bounding boxes.
        """
        if self.images_to_ram:
            image_path = annotation[0]
            image = annotation[2]
        else:
            image_path = annotation[0]
            image = cv.imread(image_path)
            
        bboxes = np.array([list(map(int, box.split(','))) for box in annotation[1]])

        if self.data_aug:
            image, bboxes = self.random_horizontal_flip(np.copy(image), np.copy(bboxes))
            image, bboxes = self.random_crop(np.copy(image), np.copy(bboxes))
            image, bboxes = self.random_translate(np.copy(image), np.copy(bboxes))

        if mAP == True: 
            return image, bboxes
        frame = Frame(np.copy(image))
        image, bboxes = ResizeWithBBox(frame).apply([self.input_sizes, self.input_sizes], np.copy(bboxes))
        return image, bboxes

    def preprocess_true_boxes(self, bboxes):
        """To prepare labels (true_boxes).

        Args:
            bboxes: a bounding boxes array.

        Returns:
            Ground true bounding boxes
        """
        label = [np.zeros((self.train_output_sizes[i], self.train_output_sizes[i], self.anchor_per_scale,
                           5 + self.num_classes)) for i in range(3)]
        bboxes_xywh = [np.zeros((self.max_bbox_per_scale, 4)) for _ in range(3)]
        bbox_count = np.zeros((3,))
        compute = YoloV3Calculus()
        for bbox in bboxes:
            bbox_coor = bbox[:4]
            bbox_class_ind = bbox[4]

            onehot = np.zeros(self.num_classes, dtype=np.float)
            onehot[bbox_class_ind] = 1.0
            uniform_distribution = np.full(self.num_classes, 1.0 / self.num_classes)
            deta = 0.01
            smooth_onehot = onehot * (1 - deta) + deta * uniform_distribution

            bbox_xywh = np.concatenate([(bbox_coor[2:] + bbox_coor[:2]) * 0.5, bbox_coor[2:] - bbox_coor[:2]], axis=-1)
            bbox_xywh_scaled = 1.0 * bbox_xywh[np.newaxis, :] / self.strides[:, np.newaxis]

            iou = []
            exist_positive = False
            for i in range(3):
                anchors_xywh = np.zeros((self.anchor_per_scale, 4))
                anchors_xywh[:, 0:2] = np.floor(bbox_xywh_scaled[i, 0:2]).astype(np.int32) + 0.5
                anchors_xywh[:, 2:4] = self.anchors[i]
                
                
                iou_scale = compute.bbox_iou(bbox_xywh_scaled[i][np.newaxis, :], anchors_xywh)
                iou.append(iou_scale)
                iou_mask = iou_scale > 0.3

                if np.any(iou_mask):
                    xind, yind = np.floor(bbox_xywh_scaled[i, 0:2]).astype(np.int32)

                    label[i][yind, xind, iou_mask, :] = 0
                    label[i][yind, xind, iou_mask, 0:4] = bbox_xywh
                    label[i][yind, xind, iou_mask, 4:5] = 1.0
                    label[i][yind, xind, iou_mask, 5:] = smooth_onehot

                    bbox_ind = int(bbox_count[i] % self.max_bbox_per_scale)
                    bboxes_xywh[i][bbox_ind, :4] = bbox_xywh
                    bbox_count[i] += 1

                    exist_positive = True

            if not exist_positive:
                best_anchor_ind = np.argmax(np.array(iou).reshape(-1), axis=-1)
                best_detect = int(best_anchor_ind / self.anchor_per_scale)
                best_anchor = int(best_anchor_ind % self.anchor_per_scale)
                xind, yind = np.floor(bbox_xywh_scaled[best_detect, 0:2]).astype(np.int32)

                label[best_detect][yind, xind, best_anchor, :] = 0
                label[best_detect][yind, xind, best_anchor, 0:4] = bbox_xywh
                label[best_detect][yind, xind, best_anchor, 4:5] = 1.0
                label[best_detect][yind, xind, best_anchor, 5:] = smooth_onehot

                bbox_ind = int(bbox_count[best_detect] % self.max_bbox_per_scale)
                bboxes_xywh[best_detect][bbox_ind, :4] = bbox_xywh
                bbox_count[best_detect] += 1
        label_sbbox, label_mbbox, label_lbbox = label
        sbboxes, mbboxes, lbboxes = bboxes_xywh
        return label_sbbox, label_mbbox, label_lbbox, sbboxes, mbboxes, lbboxes

    def __len__(self):
        """Get num of batches it will be num_samples / batch_size"""
        return self.num_batchs


import glob
import os
import cv2 as cv
import numpy as np 
import requests
import re
import errno

from py2vision.image_process.frame_decorator import Frame
from py2vision.image_process.resize import Resize
from py2vision.image_process.rotate import Rotate

class Camera():
    """An emulation of a real world camera with his relevant parameters, like camera matrix, extrinsics and intrinsics parameters. 

    Args:
        id: A string to identify our camera.
        source: When you use webcam you need to put (int) 0, if you want to use videos or streaming, you will need to put his URL or path even you can put in a video file or a image path.
    """
    def __init__(self, id, source):
        self.id = id
        self.source = source
        # define type of source
        regex = '((http|https)://)(www.)?' + '[a-zA-Z0-9@:%._\\+~#?&//=]{2,256}\\.[a-z]' + '{2,6}\\b([-a-zA-Z0-9@:%._\\+~#?&//=]*)'
        regular_exp = re.compile(regex)
        if type(self.source) is str:
            if re.search(regular_exp, self.source):
                self.type_source = 'stream'
            else:
                self.type_source = 'other'
        else:
            self.type_source = 'webcam'

    def take_photos(self, num_photos=15, save_dir="images", prefix_name="photo", rotate=0, resize=False, resize_dim=(640,480)):
        """ A simple way to take photos in console and save in a folder.

        Args:
            num_photos: Number of photos to take, 15 by default.
            save_dir: Directory name where the photos will be saved.
            prefix_name: A prefix for the names of the photos.
            rotate: rotate input image around his center, 0 grades by default.
            resize: To resize input image.
            resize_dim: resize input image dimensions (width, height), its default is (640, 480).

        Raises:
            OSError: if folder exist.
        """
        # initial directory
        cwd = os.getcwd()
        # make directory
        try:
            os.mkdir(save_dir)
            os.chdir(save_dir)
        except OSError as error:
            print(error)   
            if error.errno == errno.EEXIST:
                os.chdir(save_dir)


        if self.type_source == 'other' or self.type_source == 'webcam':
            cap = cv.VideoCapture(self.source)
            if not cap.isOpened():
                print("Cannot open camera")
                exit()

        for i in range(num_photos):
            print("Please press enter to take a photo")
            while True:
                if self.type_source == 'stream':
                    img_resp = requests.get(self.source)
                    img_arr = np.array(bytearray(img_resp.content), dtype=np.uint8)
                    frame = cv.imdecode(img_arr, -1)
                    if rotate != 0:
                        transf_frame = Frame(frame)
                        frame = Rotate(transf_frame).apply(90)
                        frame = cv.flip(frame, 0)
                    if resize:
                        transf_frame = Frame(frame)
                        frame = Resize(transf_frame).apply(resize_dim[0], resize_dim[1])
                    cv.namedWindow("streaming: {}".format(self.id), cv.WINDOW_NORMAL)
                    cv.imshow("streaming: {}".format(self.id), frame)
                elif self.type_source == 'other' or 'webcam':
                    _, frame = cap.read()
                    if rotate != 0:
                        transf_frame = Frame(frame)
                        frame = Rotate(transf_frame).apply(90)
                        frame = cv.flip(frame, -1)
                    if resize:
                        transf_frame = Frame(frame)
                        frame = Resize(transf_frame).apply(resize_dim[0], resize_dim[1])
                    cv.imshow("webcam: {}".format(self.id), frame)
                input_key = cv.waitKey(1)
                
                if input_key == 32:
                    #space pressed
                    cv.imwrite('{}_{}.png'.format(prefix_name, i + 1), frame)
                    print("saved as {}_{}.png".format(prefix_name, i + 1))
                    print("{} photos left".format(num_photos - i - 1))
                    break
                if input_key == 27:
                    #esc pressed
                    print("Escape hit, closing...")
                    break
            if input_key == 27:
                #esc pressed
                break
        cv.destroyAllWindows()
        # restore to initial directory
        os.chdir(cwd)

    def calibrate(self, images_path='', pattern_type='chessboard', pattern_size=(8,5), export_file=True, show=True, fish_eye=True):
        """ Compute camera parameters.

        Args:
            images_path: folder where is saved calibration pattern photos.
            pattern_type: It can be "circles" pattern or "chessboard" pattern (default).
            pattern_size: If pattern_type is "chessboard"  this the Number of inner corners per a chessboard row and column. But If pattern_type is "circles" this will be the number of circles per row and column. 
            export_file: To export camera parameters on xml file.
            show: if is true it show corners or centers found by calibration algorithm  at each iteration.
            fish_eye: A boolean if is true it will calibrate with cv2.fisheye.calibrate if no it'll use normal calibration, fish eye is recomended when cameras has an field of view > 160.

        Returns: 
            RMS calibration's error.

        Raises:
            OSError: If didn't find photos on images_path folder.
            ValueError: If pattern_type is different of 'chessboard' or 'circles' or when
            fish_eye isn't a boolean.
        """
        valid_patterns = ["chessboard", "circles"]
        if pattern_type not in valid_patterns:
            raise ValueError("pattern_type only can be 'chessboard' or 'circles'") 
        if type(fish_eye) != bool:
            raise ValueError("fish_eye has to be a boolean") 
        # adjust image path
        images_path, _ = os.path.splitext(images_path)
        # available formats
        formats = ['*.svg', '*.png', '*.jpg', '*.bmp', '*.jpeg', '*.raw']
        # get files names
        files_grabbed = [glob.glob((images_path + '/' + e)  if len(images_path) > 0 else e) for e in formats]
        # flatting files list
        images = [item for sublist in files_grabbed for item in sublist]
        
        # termination criteria
        if fish_eye == False:
            criteria = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 30, 0.001)
        else:
            criteria = (cv.TERM_CRITERIA_EPS+cv.TERM_CRITERIA_MAX_ITER, 30, 0.1)
        
        # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
        if fish_eye == False:
            objp = np.zeros((pattern_size[1]*pattern_size[0],3), np.float32)
            objp[:,:2] = np.mgrid[0:pattern_size[0],0:pattern_size[1]].T.reshape(-1,2)
        else:
            objp = np.zeros((pattern_size[1]*pattern_size[0], 1, 3), np.float32)
            objp[:,0, :2] = np.mgrid[0:pattern_size[0],0:pattern_size[1]].T.reshape(-1,2)
        # Arrays to store object points and image points from all the images.
        objpoints = [] # 3d point in real world space
        imgpoints = [] # 2d points in image plane.
        try:
            if len(images) == 0:
                raise OSError
            for fname in images:
                img = cv.imread(fname)
                gray = cv.cvtColor(img, cv.COLOR_BGR2GRAY)
                # Find the chess board corners
                if pattern_type == 'chessboard' and fish_eye == False:
                    ret, corners = cv.findChessboardCorners(gray, pattern_size, None)
                elif pattern_type == 'circles' and fish_eye == False:
                    ret, corners = cv.findCirclesGrid(gray, pattern_size, None)
                elif pattern_type == 'chessboard' and fish_eye == True:
                    ret, corners = cv.findChessboardCorners(gray, pattern_size, flags=cv.CALIB_CB_ADAPTIVE_THRESH+cv.CALIB_CB_FAST_CHECK+cv.CALIB_CB_NORMALIZE_IMAGE)
                elif pattern_type == 'circles' and fish_eye == True:
                    ret, corners = cv.findCirclesGrid(gray, pattern_size, cv.CALIB_CB_ADAPTIVE_THRESH+cv.CALIB_CB_FAST_CHECK+cv.CALIB_CB_NORMALIZE_IMAGE)
                # If found, add object points, image points (after refining them)
                if ret == True:
                    objpoints.append(objp)
                    if fish_eye == False:   
                        corners = cv.cornerSubPix(gray,corners, (11,11), (-1,-1), criteria)
                    else:
                        corners = cv.cornerSubPix(gray, corners, (3,3), (-1,-1), criteria)
                    imgpoints.append(corners)

                if show:
                    # Draw and display the corners
                    cv.drawChessboardCorners(img, pattern_size, corners, ret)
                    cv.namedWindow(fname, cv.WINDOW_NORMAL)
                    cv.imshow(fname, img)
                    cv.waitKey(3000)
                    cv.destroyAllWindows()

            if fish_eye:
                flags = cv.fisheye.CALIB_RECOMPUTE_EXTRINSIC + cv.fisheye.CALIB_FIX_SKEW
                k = np.zeros((3, 3))
                d = np.zeros((4, 1))
                rvecs = [np.zeros((1, 1, 3), dtype=np.float64) for i in range(len(objpoints))]
                tvecs = [np.zeros((1, 1, 3), dtype=np.float64) for i in range(len(objpoints))]

            if fish_eye == False:
                error, self.matrix, self.dist_coeffs, self.rvecs, self.tvecs = cv.calibrateCamera(objpoints, imgpoints, gray.shape[::-1], None, None)
                height, width = img.shape[:2]
                self.matrix, self.roi = cv.getOptimalNewCameraMatrix(self.matrix, self.dist_coeffs, (width, height), 1, (width, height))
               
            else:
                error, self.matrix, self.dist_coeffs, self.rvecs, self.tvecs = cv.fisheye.calibrate(objpoints, imgpoints, gray.shape[::-1], k, d, rvecs, tvecs, flags, (cv.TERM_CRITERIA_EPS+cv.TERM_CRITERIA_MAX_ITER, 30, 1e-6))
            print("camera error {}".format(error))
            # export an xml with camera parameters
            if export_file:
                print("Saving parameters!")
                cv_file = cv.FileStorage('{}_parameters.xml'.format(self.id), cv.FILE_STORAGE_WRITE)
                cv_file.write('camera_mtx', self.matrix)
                cv_file.write('distortion_coeff', self.dist_coeffs)
                for i, rvec in enumerate(self.rvecs):
                    cv_file.write('rotation_mtx_col_{}'.format(i), rvec) 
                num_rvecs = i + 1
                cv_file.write('num_rvecs', num_rvecs)
                for i, tvec in enumerate(self.tvecs):
                    cv_file.write('translation_mtx_col_{}'.format(i), tvec)
                num_tvecs = i + 1
                cv_file.write('num_tvecs', num_tvecs)
                if hasattr(self, "roi"):
                    cv_file.write('ROI', self.roi)
                cv_file.release()
                print("Saved as {}_parameters.xml".format(self.id))
        except OSError:
            print("Could not find any image in {}".format(images_path))

        return error

    def get_parameters(self, path):
        """ Loads camera parameters from a xml file
        Args: 
            path: a path where is saved xml file. (Note: if you don't have any parameters you can use calibrate method first and then pass true in export file argument)
        Raises:
            OSError: when path is wrong.
         """

        # file storage read
        cv_file = cv.FileStorage(path, cv.FILE_STORAGE_READ)
        if not cv_file.isOpened():
            raise OSError("Looks like {} doesn't exist!".format(path))
        self.matrix = cv_file.getNode('camera_mtx').mat()
        self.dist_coeffs = cv_file.getNode('distortion_coeff').mat()
        num_rvecs = int(cv_file.getNode('num_rvecs').real())
        self.rvecs = [cv_file.getNode('rotation_mtx_col_{}'.format(i)).mat() for i in range(num_rvecs)]
        num_tvecs = int(cv_file.getNode('num_tvecs').real())
        self.tvecs = [cv_file.getNode('translation_mtx_col_{}'.format(i)).mat() for i in range(num_tvecs)]
        if hasattr(self, "roi"):
            self.roi = tuple(cv_file.getNode('ROI').mat().astype(int).reshape(1, -1)[0])

        cv_file.release()



import tensorflow as tf
from tensorflow.keras.layers import BatchNormalization

class BatchNormalization(BatchNormalization):
    """ A modified batch normalization layer.

    Args:
        x: input of the layer.
        training: it's necessary to know when freeze the layer weights.

    Returns:
        A batchNormalization layer.
    """
    def call(self, x, training=False):
        if not training:
            training = tf.constant(False)
        training = tf.logical_and(training, self.trainable)
        return super().call(x, training)


import os

def label_map(labels, dst_path=None, name='label_map'):
    """An easy way to convert classes names and ids to a .pbtxt file compatible with tensorflow models API.
    
    Args:
        labels: a list, which each element is a dictionary with two keys 'name' and 'id'.
        dst_path: a path where the file will be saved.
        name: the name of the file, with which it will save the .pbtxt

    Returns:
        a string where the file was saved.

    Raises:
        Exception: When someone element in internal dictionaries have another keys different of name and id.
        TypeError: when labels aren't list type
        ValueError: when labels are empty
    """
    if type(labels) != list: raise TypeError('labels should be a list of dictionaries!')
    if not labels: raise ValueError('labels are empty')
    if dst_path == None:
        dst_path = os.getcwd()
    file_path = dst_path +  '/{}.pbtxt'.format(name)
    allowed_keys = ['name', 'id']
    with open(file_path, 'w') as f:
        for label in labels:
            if allowed_keys == list(label.keys()):
                f.write('item { \n')
                f.write('\tname:\'{}\'\n'.format(label['name']))
                f.write('\tid:{}\n'.format(label['id']))
                f.write('}\n')
            else: 
                os.remove(file_path)
                raise Exception("Your keys should be only 'name' and 'id'")
        print('file saved in: {}'.format(file_path))
    
    return file_path
            
def index2class(index, classes: list):
    """Convert index (int) to class name (string)"""
    return classes[index]


def class2index(class_name, classes: list):
    """Convert class name (string) to index (int)"""
    return classes.index(class_name)

def class_text_to_int(row_label, label_map_dict: dict):
    return label_map_dict[row_label]

def read_class_names(class_file_name):
    """loads class name from a file to a dict"""
    names = {}
    with open(class_file_name, 'r') as data:
        for ID, name in enumerate(data):
            names[ID] = name.strip('\n')
    return names